/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Authentication stuff.
 */

#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/DialogS.h>
#include <Xm/List.h>

#include <stdio.h>
#include <pwd.h>
#include <malloc.h>
#include <string.h>
#include <sys/types.h>
#include <memory.h>
#include <unistd.h>
#include <sys/time.h>
#include <ctype.h>

#include <portable.h>
#include <lber.h>
#include <ldap.h>
#include <disptmpl.h>

#ifdef KERBEROS
#include <krb.h>
#endif /* KERBEROS */

#include "dtypes.h"
#include "auth.h"
#include "globals.h"
#include "util.h"
#include "dialog.h"
#include "find.h"
#include "search.h"
#include "help.h"

/* Forward declarations */
#ifndef NEEDPROTOS
static void	get_identity();
static void	create_authAsDialog();
static void	do_authRead();
static void	check_authRead();
static void	cancel_authRead();
static void	check_passwd();
static void	init_auth();
static void	get_password();
void	start_auth();
#else /* !NEEDPROTOS */
static void	get_identity();
static void	create_authAsDialog(Widget w);
static void	do_authRead(Widget w, LDAP *ld, char *dn, Widget status);
static void	check_authRead(struct searchData *f);
static void	cancel_authRead(Widget w, XtPointer client_data,
			XmAnyCallbackStruct *cbs);
static void	check_passwd(Widget w, XtPointer client_data,
			XmTextVerifyCallbackStruct *cbs);
static void	init_auth();
static void	get_password();
void		start_auth(Widget w, XtPointer callback, XtPointer call_data,
			Boolean explicit);
#endif /* !NEEDPROTOS */

#ifdef KERBEROS
#ifndef NEEDPROTOS
int		valid_tgt();
static int	kinit();
static void	afs_transarc_StringToKey();
static void	afs_cmu_StringToKey();
static void	afs_string_to_key();
extern char	*crypt();
#else /* !NEEDPROTOS */
int		valid_tgt(char *kname);
static int	kinit(char *kname);
static void	afs_transarc_StringToKey(char *str, char *cell,
			des_cblock key);
static void	afs_cmu_StringToKey(char *str, char *cell, des_cblock key);
static void	afs_string_to_key(char *str, des_cblock key, char *cell);
extern char	*crypt(char *key, char *salt);
extern int	krb_get_cred();
extern int	kname_parse();
extern int	des_string_to_key();
extern int	krb_get_in_tkt();
extern int	des_key_sched();
#endif /* !NEEDPROTOS */
#endif /* KERBEROS */

#ifdef KERBEROS
static char	*krbNameAttrs[] = {"krbName", "objectClass", NULL};
static char	*krbNameDisAttrs[] = {"title", NULL};
#else /* KERBEROS */
static char	*krbNameAttrs[] = {NULL};
static char	*krbNameDisAttrs[] = {NULL};
#endif /* KERBEROS */

char		*passwd;
authStruct	*auth = NULL;
static char	auth_hint[1024] = "\0";


/*
 * Rebind as NULL.  Note: does not destroy kerberos tickets.
 */
void
do_authDiscard(w, item_no, cbs)
Widget			w;
int			item_no;
XmAnyCallbackStruct	*cbs;
{
    ldap_simple_bind_s(ldp, appConfig.ldapBindDN,
	    NULL);
    bound_as[0] = '\0';
    sticky_bind = False;
    updateBoundAsText(bound_as);
}




/*
 * Start to authenticate.  do_auth() is designed to be invoked as
 * a callback.  To directly start authenticating from another chunk
 * of code, use start_auth().
 */
void
do_auth(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    /*
     * We assume we're being invoked as a result of user choosing
     * the "authenticate" menu (or its shortcut) and therefore
     * use toplevel as the parent.
     */
    start_auth(toplevel, NULL, NULL, True);
}




/*
 * Start to authenticate.  Allocate the global authStruct structure,
 * which gets filled in as we go through the authentication process.
 *
 * Note: "sticky" authentication means that when saving an entry, even
 * if we don't appear to be bound as something which will allow a write
 * to succeed, use the "sticky" credentials.  Typically used when
 * binding as "Manager".
 */
void
start_auth(w, callback, call_data, explicit)
Widget		w;
XtPointer	callback;	/* fn to call when we're all done */
XtPointer	call_data;	/* data to pass to fn */
Boolean		explicit;	/* if nonzero, make auth sticky */
{
    int		item, i;
    char	**items, **p;

    init_auth();
    auth->callback = (void *) callback;
    auth->call_data = call_data;
    auth->parent = GetTopShell(w);
    auth->sticky = explicit;
    get_identity();

    if (auth->numDNs == 0 || explicit) {
	/* 
	 * Don't know which entry to bind as yet, or are
	 * explicitly authenticating.
	 */
	create_authAsDialog(auth->parent);
	return;
    } else if (auth->numDNs > 1) {
	/* We have more than one possible DN - present a list */
	/* make a selection dialog with RDNs listed */
	items = (char **) ch_malloc((auth->numDNs + 1) * sizeof(char *));
	for (i = 0; i < auth->numDNs; i++) {
	    p = ldap_explode_dn(auth->dn_list[i], 1);
	    items[i] = strdup(p[0]);
	    ldap_value_free(p);
	}
	items[auth->numDNs] = NULL;
	item = modalSelectionDlg(auth->parent, "authDNDlg",
		items);
	if (item == -1) {
	    goto clean;
	}
	auth->dn = auth->dn_list[item];
clean:	for (i = 0; items[i] != NULL; i++) {
	    free(items[i]);
	}
	free(items);
    } else {
	/* exactly one DN */
	auth->dn = auth->dn_list[0];
    }
#ifdef KERBEROS
    if (auth->dn) {
	/*
	 * Get the krbName attribute, so we can see if user has some
	 * reasonable kerberos tickets.
	 */
	do_authRead(auth->parent, ldp, auth->dn, NULL);
    }
#else /* KERBEROS */
    /* Don't need to do any reads in simple bind case */
    auth->authTypes = 
	    (authTypeStruct *) ch_malloc(2 * sizeof(authTypeStruct));
    memset(auth->authTypes, 0, 2 * sizeof(authTypeStruct));
    auth->activeAuth = &(auth->authTypes[0]);
    auth->activeAuth->method = LDAP_AUTH_SIMPLE;
    p = ldap_explode_dn(auth->dn_list[0], 1);
    auth->activeAuth->data = strdup(p[0]);
    ldap_value_free(p);
    auth->authTypes[1].data = NULL;
    get_password(w, auth);
#endif /* KERBEROS */

}



/*
 * Initiate a kerberos bind operation.  Sets up the appropriate fields
 * in the auth struct and calls start_async_bind().
 */
void
auth_kerberos(w, krbname, callback, d)
Widget			w;
char			*krbname;
XtPointer		callback;
displayData		*d;
{
    int		rc;
    static char	errstr[256];

    init_auth();
    auth->callback = (void *) callback;
    auth->call_data = (void *) d;
    auth->parent = GetTopShell(w);
    auth->numDNs = 1;
    auth->dn = d->di_entryData->en_dn;
    auth->authTypes = (authTypeStruct *) ch_malloc( 2 * sizeof(authTypeStruct));
    auth->authTypes[1].data = NULL;
    auth->activeAuth = auth->activeAuth = &(auth->authTypes[0]);
    auth->activeAuth->method = LDAP_AUTH_KRBV41;
    auth->activeAuth->data = strdup(krbname);
    rc = start_async_bind();
    if (rc < 0) {
	sprintf(errstr, "Error (krb42): %s", ldap_err2string(ldp->ld_errno));
	errorDlg(GetTopShell(w), errstr);
    }
}





/*
 * Try to figure out who we should bind as.  Algorithm:
 * Get the most recently-created entry window and use
 * its DN.
 * If no entries are on-screen, return the user's login ID
 * (obtained from the passwd file).
 *
 * Once we've decided who to try to bind as, fill in the global
 * "auth_hint".  When the "authenticate as" dialog comes up, it'll
 * be filled in with the contents of the auth_hint string.
 */
static void
get_identity()
{
struct passwd		*pwent;
struct berval		**bv;
displayAttrData		*dad;
int			i;

    if (latestDisplayDlg != NULL) {
	strcpy(auth_hint, latestDisplayDlg->di_entryData->en_dn);
	for (dad = latestDisplayDlg->di_displayAttrData;
		dad != NULL; dad = dad->dad_next) {
	    if ((dad->dad_attrData->ad_attrname != NULL) && 
		    !strcasecmp(dad->dad_attrData->ad_attrname, "owner")) {
		break;
	    }
	}
	if (dad == NULL) {
	    auth->dn_list = (char **) ch_malloc(2 * sizeof(char *));
	    auth->dn_list[0] = strdup(latestDisplayDlg->di_entryData->en_dn);
	    auth->dn_list[1] = NULL;
	    auth->dn = auth->dn_list[0];
	    auth->numDNs = 1;
	} else {
	    bv = dad->dad_attrData->ad_vals;
	    for (i = 0; bv[i] != NULL; i++);
	    auth->dn_list = (char **) ch_malloc((i + 1) * sizeof(char *));
	    memset(auth->dn_list, 0, (i + 1) * sizeof(char *));
	    for (i = 0; bv[i] != NULL; i++) {
		auth->dn_list[i] = strdup(bv[i]->bv_val);
	    }
	    auth->dn_list[i] = NULL;
	    auth->numDNs = i;
	}
    } else {
	if ((pwent = getpwuid(getuid())) != NULL) {
	    strcpy(auth_hint, pwent->pw_name);
	}
    }
    return;
}





/*
 * Check to see if the asynchronous bind operation has completed.
 */
static void
check_async_bind(a)
authStruct	*a;
{
    int			rc;
    int			err;
    LDAPMessage		*result;
    char		*bound_as_ufn;
    static char		errstr[256];

    rc = ldap_result(ldp, a->msgid, 1, zerotime, &result);
    switch (rc) {
    case 0:
	/* search not yet complete */
	a->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
		check_async_bind, a);
	return;
	break;
    case -1:
	a->ldaprc = ldp->ld_errno;
	ldap_msgfree(result);
	if ((reconnect_ldap(a->ldaprc)) == 0) {
	    /* reconnect successful, retry */
	    err = start_async_bind();
	    if (err < 0) {
		sprintf(errstr, "Error: %s", ldap_err2string(ldp->ld_errno));
		errorDlg(GetTopShell(auth->parent), errstr);
	    }
	    return;
	}
	break;
    default:
	/*
	 * Bind operation finished.  If we're doing simple authentication,
	 * then we're done.  If kerberos, we need to start the second
	 * half of the bind operation (kerberos bind is done in two steps,
	 * one to authenticate to the ldapserver and one to authenticate
	 * to the DSA).
	 */
	if (a->workingDlg != NULL) {
	    XtDestroyWidget(a->workingDlg);
	    a->workingDlg = NULL;
	}
	a->ldaprc = ldap_result2error(ldp, result, 0);
	ldap_msgfree(result);
	if ((a->ldaprc == LDAP_SUCCESS)
	   || (a->ldaprc == LDAP_SIZELIMIT_EXCEEDED)
	   || (a->ldaprc == LDAP_SIZELIMIT_EXCEEDED)) {
	    /* bind was successful */
	    if (a->activeAuth->method == LDAP_AUTH_KRBV41) {
		/* Start second half of kerberos bind operation */
		a->activeAuth->method = LDAP_AUTH_KRBV42;
		err = start_async_bind();
		if (err < 0) {
		    sprintf(errstr, "Error: %s",
			    ldap_err2string(ldp->ld_errno));
		    errorDlg(GetTopShell(auth->parent), errstr);
		}
		return;
	    } else {
		/* Handle simple auth case */
		strcpy(bound_as, a->dn ? a->dn : "");
		sticky_bind = auth->sticky;
		bound_as_ufn = a->dn ? ldap_dn2ufn(a->dn) : NULL;
		updateBoundAsText(bound_as_ufn);
		if (auth->callback) {
		    auth->callback(auth->call_data);
		}
		if (bound_as_ufn != NULL) {
		    free(bound_as_ufn);
		}
	    }
	} else {
	    /* Bind operation failed.  Generate a useful error message */
	    if (a->ldaprc == LDAP_NO_SUCH_ATTRIBUTE) {
		errorDlg(GetTopShell(auth->parent), "Entry has no password");
	    } else if (a->ldaprc == LDAP_INVALID_CREDENTIALS) {
		errorDlg(GetTopShell(auth->parent), "Incorrect password");
	    } else {
		sprintf(errstr, "Error: %s",
			ldap_err2string(ldap_result2error(ldp, result, 0)));
		errorDlg(GetTopShell(auth->parent), errstr);
	    }
	    /*
	     * If authentication failed, we need to rebind as nobody, or
	     * subsequent LDAP operations will fail.
	     */
	    err = ldap_simple_bind_s(ldp, appConfig.ldapBindDN, NULL);
	    bound_as[0] = '\0';
	    sticky_bind = False;
	    updateBoundAsText(bound_as);
	    if (err != LDAP_SUCCESS) {
		errorDlg(GetTopShell(auth->parent), "Cannot rebind");
	    }
	}
    }
}




/*
 * Initiate an asynchronous bind operation.  Check for completion
 * in check_async_bind().
 */
int
start_async_bind()
{
    if (auth == NULL) {
	/* Just bind as NULL */
	init_auth();
	auth->authTypes = (authTypeStruct *) ch_malloc(sizeof (authTypeStruct));
	auth->authTypes->method = LDAP_AUTH_SIMPLE;
	auth->authTypes->data =  NULL;
	auth->activeAuth = auth->authTypes;
	auth->sticky = False;
	passwd = NULL;
    }

    auth->msgid = ldap_bind(ldp, auth->dn, passwd, auth->activeAuth->method);
    if (passwd != NULL) {
	/* Be paranoid and zero out the password */
	memset(passwd, 0, strlen(passwd));
	XtFree(passwd);
	passwd = NULL;
    }

    if (auth->msgid < 0) {
	return(auth->msgid);	/* error */
    }

    auth->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
			          check_async_bind, auth);
    return(0);
}




/*
 * Decide which authentication method to use, and initiate a bind
 * operation.
 */
void
do_bind(w, a, cbs)
Widget			w;
authStruct		*a;
XmAnyCallbackStruct	*cbs;
{
    authTypeStruct	*aauth = auth->activeAuth;
    int			rc;
    static char		errstr[256];

    switch(aauth->method) {
    case LDAP_AUTH_SIMPLE:
	rc = start_async_bind();
	if (rc < 0) {
	    sprintf(errstr, "Error: %s", ldap_err2string(ldp->ld_errno));
	    errorDlg(GetTopShell(w), errstr);
	}
	break;
#ifdef KERBEROS
    case LDAP_AUTH_KRBV41:
	/* get a Kerberos TGT */
	if (kinit(aauth->data)) {
	    return;
	}
	/* bind here */
	rc = start_async_bind();
#endif /* KERBEROS */
    default:
	break;
    }
    if (passwd != NULL) {
	/* zero password */
	memset(passwd, 0, strlen(passwd));
	XtFree(passwd);
    }
}




/*
 * Prompt user for a password.
 */
static void
get_password(parent, a)
Widget		parent;
authStruct	*a;
{
    Widget	dlg, text;
    Arg		args[3];
    int		n;

    n = 0;
    XtSetArg(args[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
    dlg = XmCreatePromptDialog(parent, "passwdDialog", args, n);
    text = XmSelectionBoxGetChild(dlg, XmDIALOG_TEXT);
    XtAddCallback(text, XmNmodifyVerifyCallback, (XtCallbackProc) check_passwd,
	    NULL);
    XtAddCallback(dlg, XmNokCallback, do_bind, a);
    XtManageChild(dlg);
}
	



/*
 * Convert characters typed in a text widget to '*'.  From O'Reilly
 * Volume 6A.
 */
static void
check_passwd(w, client_data, cbs)
Widget				w;
XtPointer			client_data;
XmTextVerifyCallbackStruct	*cbs;
{
    char	*new;
    int		len;


    if (cbs->reason == XmCR_ACTIVATE) {
	return;
    }

    if (cbs->text->ptr == NULL) { /* backspace */
	cbs->endPos = strlen(passwd); /* delete from here to end */
	if (cbs->endPos <= 0) {
	    return; /* catch null passwd */
	}
	passwd[cbs->startPos] = 0; /* backspace--terminate */
	return;
    }

    if (cbs->text->length > 1) {
	cbs->doit = False;
	return;
    }

    new = XtMalloc(cbs->endPos + 2); /* new char + NULL terminator */
    if (passwd) {
	strcpy(new, passwd);
	XtFree(passwd);
    } else {
	new[0] = '\0';
    }

    passwd = new;
    strncat(passwd, cbs->text->ptr, cbs->text->length);
    passwd[cbs->endPos + cbs->text->length] = 0;

    for (len = 0; len < cbs->text->length; len++) {
	cbs->text->ptr[len] = '*';
    }
}





/*
 * Check to see if the read operation for authentication has finished.
 */
static void
check_authRead(f)
struct searchData *f;
{
    int			rc;
    LDAPMessage		*result = NULL, *e;
    char		buf[100];
    static char		errstr[256];
    char		*krbNameWithTickets, **krbNames = NULL;
    char		**aitems;
    char		**xdn, *dn;
    int			i, resp;
    authTypeStruct	*authTypes;
    int			numAuthMethods = 1;
    Boolean		have_tickets;


    /* See if the ldap_search() has completed */
    rc =  ldap_result(ldp, f->msgid, 1, zerotime, &result);
    switch (rc) {
    case 0:
	/* search not yet complete */
	f->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
		(XtTimerCallbackProc) check_authRead, f);
	if (result != NULL) {
	    ldap_msgfree(result);
	}
	return;
	break;
    case -1:
	/* fatal error */
	f->ldaprc = ldp->ld_errno;
	if (result != NULL) {
	    ldap_msgfree(result);
	}
	if ((reconnect_ldap(f->ldaprc)) == 0) {
	    /* reconnect successful, retry */
	    do_authRead(f->parent, ldp, f->search_string, f->status);
	    return;
	}
	break;
    default:
	/*
	 * Read is finished.  Collect a list of all the possible authentication
	 * methods the user can choose.  If there's only one, proceed to that
	 * authentication method, otherwise present a scrolling list of
	 * authentication methods and allow the user to choose one.
	 */
	/* If there's a working dialog on-screen, destroy it */
	if (f->workingDlg != NULL) {
	    XtDestroyWidget(f->workingDlg);
	    f->workingDlg = NULL;
	}
	f->ldaprc = ldap_result2error(ldp, result, 0);
	if ((f->ldaprc == LDAP_SUCCESS)
	   || (f->ldaprc == LDAP_SIZELIMIT_EXCEEDED)
	   || (f->ldaprc == LDAP_SIZELIMIT_EXCEEDED)) {
	    /* search ok, display results */
	    numAuthMethods = 1;
	    e = ldap_first_entry(ldp, result);
#ifdef KERBEROS
	    krbNames = ldap_get_values(ldp, e, "krbName");
	    /*
	     * If it looks like we've got valid kerberos
	     * tickets, don't bother presenting the user with
	     * the authMethod dialog.
	     */
	    have_tickets = False;
	    for (i = 0; krbNames && krbNames[i] != NULL; i++) {
		if (valid_tgt(krbNames[i])) {
		    have_tickets = True;
		    krbNameWithTickets = krbNames[i];
		    break;
		}
	    }
	    if (have_tickets) {
		numAuthMethods = 1;
	    } else {
		numAuthMethods += ldap_count_values(krbNames);
	    }
#endif /* KERBEROS */
	    authTypes = (authTypeStruct *) ch_malloc((numAuthMethods + 1) *
		    sizeof(authTypeStruct));
	    authTypes[numAuthMethods].data = NULL;
	    aitems = (char **) ch_malloc((numAuthMethods + 1) * sizeof(char *));
	    dn = ldap_get_dn(ldp, e);
	    xdn = ldap_explode_dn(dn, 1);
	    auth->dn = dn;
	    auth->workingDlg = NULL;
	    auth->authTypes = authTypes;
#ifdef KERBEROS
	    if (have_tickets) {
		authTypes[0].method = LDAP_AUTH_KRBV41;
		authTypes[0].data = strdup(krbNameWithTickets);
		aitems[0] = strdup("");
		auth->activeAuth = &authTypes[0];
		rc = start_async_bind();
		if (rc < 0) {
		    sprintf(errstr, "Error (krb42): %s",
			    ldap_err2string(ldp->ld_errno));
		    errorDlg(f->parent, errstr);
		}
		goto auth_cleanup;
	    }
#endif /* KERBEROS */
	    authTypes[0].method = LDAP_AUTH_SIMPLE;
	    authTypes[0].data = "";
	    sprintf(buf, "X.500 Password: %s", xdn[0]);
	    aitems[0] = strdup(buf);
#ifdef KERBEROS
	    for (i = 0; krbNames && krbNames[i] != NULL; i++) {
		authTypes[i + 1].method = LDAP_AUTH_KRBV41;
		authTypes[i + 1].data = strdup(krbNames[i]);
		sprintf(buf, "Kerberos/Uniqname: %s", krbNames[i]);
		aitems[i + 1] = strdup(buf);
	    }
#endif /* KERBEROS */
	    aitems[numAuthMethods] = NULL;
	    if (numAuthMethods > 1) {
		resp = modalSelectionDlg(f->parent, "authTypeDlg", aitems);
		if (resp == -1) {
		    goto auth_cleanup;
		}
	    } else {
		resp = 0;
	    }
	    auth->activeAuth = &authTypes[resp];
	    switch (authTypes[resp].method) {
	    case LDAP_AUTH_SIMPLE:
		get_password(f->parent, auth); /* XXX global */
		break;
#ifdef KERBEROS
	    case LDAP_AUTH_KRBV41:
		if (valid_tgt(authTypes[resp].data)) {
		    rc = start_async_bind();
		    if (rc < 0) {
			sprintf(errstr, "Error (krb42): %s",
				ldap_err2string(ldp->ld_errno));
			errorDlg(XtParent(f->parent), errstr);
		    }
		} else {
		    get_password(f->parent, auth);
		}
		    break;
#endif /* KERBEROS */
	    default:
		break;
	    }
auth_cleanup:	for (i = 0; i < numAuthMethods; i++) {
		free(aitems[i]);
	    }
	    free(aitems);
	    free_searchdata(&f);
	    ldap_value_free(xdn);
	    if (krbNames != NULL) {
		ldap_value_free(krbNames);
	    }	
	    if (result != NULL) {
		ldap_msgfree(result);
	    }
	    return;
	}
    }
    /* some kind of an error occurred */
    sprintf(errstr, "Error while searching: %s", ldap_err2string(f->ldaprc));
    errorDlg(f->parent, errstr);
    return;
}
			




/*
 * Initiate a read operation to get the krbName attributes.
 */
static void
do_authRead(w, ld, dn, status)
Widget  w;
LDAP    *ld;
char    *dn;
Widget  status;
{
    struct searchData	*f;

    /* set up our per-search struct */
    if ((f = get_searchdata()) == NULL) {
	fprintf(stderr, "out of memory in do_read\n");
	exit(1);
    }
    f->searchbase = strdup(search_base);
    f->workingDlg = (Widget) NULL;
    f->parent = auth->parent;
    f->status = status;
    f->start_time = time(NULL);
    f->etype = NULL;
    f->search_string = strdup(dn);
    /* XtFree(dn); */
    f->objcllist = NULL;
    f->msgid = ldap_search(ld, f->search_string, LDAP_SCOPE_BASE,
	    "objectclass=*", krbNameAttrs, 0);
    /* Pop up the working dialog */
    f->workingDlg = create_working_dialog(w, "Reading...", "Cancel", NULL,
	    NULL, (XtCallbackProc) cancel_authRead, NULL, NULL, f, w);
    /* Set up an Xt App timeout function to handle the result */
    f->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL, 
	    (XtTimerCallbackProc) check_authRead, f);
}




/*
 * Cancel the read operation
 */
static void
cancel_authRead(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    struct searchData	*f;
    XmString		s;

    f = (searchData *) client_data;

    /* abandon the current LDAP operation */
    ldap_abandon(ldp, f->msgid);

    if (f->status) {
	s = XmStringCreateLtoR("Read cancelled", XmSTRING_DEFAULT_CHARSET);
	XtVaSetValues(f->status, XmNlabelString, s, NULL);
	XmStringFree(s);
    }
	
    XtRemoveTimeOut(f->timer);

    /* Free the searchData struct */
    free_searchdata(&f);
}





/*
 * Create the "authenticate as:" dialog.
 */
static void
create_authAsDialog(w)
Widget		w;
{

    char	*resp;

    resp = modalPromptDialog(GetTopShell(w), "authAsDialog", auth_hint,
	    doHelpDlg, HLP_TOPIC_AUTHENTICATE);
    if (strlen(resp) == 0) {
	/* cancel */
	return;
    }
    if (is_dn(resp)) {
	do_authRead(w, ldp, resp, NULL);
    } else {
	do_search(GetTopShell(w), resp, krbNameAttrs, krbNameDisAttrs,
		appConfig.authFilter, NULL, "People", do_authRead, "authselDlg",
		0L, NULL);
    }
    XtFree(resp);
}



/*
 * Check to see if the user has a valid Kerberos
 * Ticket-Granting-Ticket (TGT).
 */
#ifdef KERBEROS
#define	FIVEMINS	5 * 60
#define	TGT		"krbtgt"
int
valid_tgt(kname)
char	*kname;
{
    char	name[ANAME_SZ], inst[INST_SZ], realm[REALM_SZ];
    CREDENTIALS	cred;

    memset(realm, 0, REALM_SZ);
    if (kname_parse(name, inst, realm, kname) != KSUCCESS) {
	return(0);
    }

    /*
     * realm must be uppercase for krb_ routines
     */
    str2upper(realm);

    /*
     * check ticket file for a valid ticket granting ticket
     * my check is: have ticket granting ticket and it is good for
     * at least 5 more minutes
     */
    if (krb_get_cred(TGT, realm, realm, &cred) == KSUCCESS) { 
	if (time(0) + FIVEMINS < cred.issue_date +
		(u_char)cred.lifetime * FIVEMINS) {
	    if (!strcmp(cred.pname, name)) {
		return(1);
	    }
	}
    }
    return(0);
}



/*
 * Dummy routine - we do our own encoding by hand.
 */
static int
use_given_key(user, inst, realm, pw, key)
char *user, *inst, *realm, *pw;
C_Block key;
{
    memcpy (key, pw, sizeof(des_cblock));
    return(0);
}



/*
 * Get a kerberos TGT for "kname" (principal@realm)
 */
static int
kinit(kname)
char	*kname;
{
	int		rc;
	char		name[ANAME_SZ], inst[INST_SZ];
	char		cell[REALM_SZ], realm[REALM_SZ];
	char		errstr[512];
	char		buf[1024];
	des_cblock	key;
	int		i;

    if (kname_parse(name, inst, realm, kname) != KSUCCESS) {
	return( -1 );
    }
    strcpy(cell, realm);

    /*
     * Try to handle both MIT and AFS kerberos transparently.
     * MIT first.
     */
    /* upcase the cell */
    str2upper(realm);
    des_string_to_key(passwd, key);
    rc = krb_get_in_tkt(name, inst, realm, TGT, realm, DEFAULT_TKT_LIFE,
	    use_given_key, NULL, key);
    if (rc == KSUCCESS) {
	memset(key, 0, sizeof(key));
	if (krb_get_cred("krbtgt", realm, realm, buf)) {
	    return(-1);
	}
	memset(buf, 0, sizeof(buf));
	return(0);
    }

    /*
     * MIT kerberos didn't work, so try AFS kerberos.
     */
    for (i = 0; cell[i]; i++) {
	if (isupper(cell[i])) cell[i] = tolower(cell[i]);
    }
    afs_string_to_key(passwd, key, cell);
    rc = krb_get_in_tkt(name, inst, realm, TGT, realm, DEFAULT_TKT_LIFE,
	    use_given_key, NULL, key);
    memset(key, 0, sizeof(key));
    if (rc == KSUCCESS) {
	memset(key, 0, sizeof(key));
	if (krb_get_cred("krbtgt", realm, realm, buf)) {
	    return(-1);
	}
	memset(buf, 0, sizeof(buf));
	return(0);
    }

    switch (rc) {
    case SKDC_CANT:
	sprintf(errstr, "Can't contact Kerberos server for realm \"%s\"\n" ,
		realm );
	errorDlg(toplevel, errstr);
	break;
    default:
	sprintf(errstr, "%s: %s\n", name, krb_err_txt[rc]);
	errorDlg(toplevel, errstr);
	break;
    }
    return( -1 );
}


/* andrewstk.c -- afs string to key function
 *
 * Code taken from AuthMan from University of Michigan
 */

extern void des_fixup_key_parity();
extern unsigned long des_cbc_cksum();

/* forward declarations */

/* This defines the Andrew string_to_key function.  It accepts a password
 * string as input and converts its via a one-way encryption algorithm to a DES
 * encryption key.  It is compatible with the original Andrew authentication
 * service password database.
 */

static void
afs_cmu_StringToKey (str, cell, key)
char		*str;
char		*cell;                  /* cell for password */
des_cblock	key;
{
    char  password[8+1];                /* crypt is limited to 8 chars anyway */
    int   i;
    int   passlen;

    memset(key, 0, sizeof(des_cblock));
    memset(password, 0, sizeof(password));

    strncpy(password, cell, 8);
    passlen = strlen(str);
    if (passlen > 8) {
	passlen = 8;
    }

    for (i = 0; i < passlen; i++) {
        password[i] = str[i] ^ cell[i];
    }

    for (i = 0; i < 8; i++) {
        if (password[i] == '\0') {
	    password[i] = 'X';
	}
    }

    /* 
     * crypt only considers the first 8 characters of password but for some
     * reason returns eleven characters of result (plus the two salt chars).
     */
    strncpy((void *)key, crypt(password, "#~") + 2, sizeof(des_cblock));

    /* 
     * parity is inserted into the LSB so leftshift each byte up one bit.  This
     * allows ascii characters with a zero MSB to retain as much significance
     * as possible. */
    {
	char *keybytes = (char *)key;
	unsigned int temp;

	for (i = 0; i < 8; i++) {
	    temp = (unsigned int) keybytes[i];
	    keybytes[i] = (unsigned char) (temp << 1);
	}
    }
    des_fixup_key_parity (key);
}

static void
afs_transarc_StringToKey (str, cell, key)
char		*str;
char		*cell;                  /* cell for password */
des_cblock	key;
{
    des_key_schedule schedule;
    char temp_key[8];
    char ivec[8];
    char password[BUFSIZ];
    int  passlen;

    strncpy(password, str, sizeof(password));
    if ((passlen = strlen (password)) < sizeof(password) - 1) {
        strncat(password, cell, sizeof(password) - passlen);
    }
    if ((passlen = strlen(password)) > sizeof(password)) {
	passlen = sizeof(password);
    }

    memcpy(ivec, "kerberos", 8);
    memcpy(temp_key, "kerberos", 8);
    des_fixup_key_parity ((void *)temp_key);
    des_key_sched (temp_key, schedule);
    des_cbc_cksum (password, ivec, passlen, schedule, ivec);

    memcpy(temp_key, ivec, 8);
    des_fixup_key_parity ((void *)temp_key);
    des_key_sched (temp_key, schedule);
    des_cbc_cksum (password, (void *)key, passlen, schedule, ivec);

    des_fixup_key_parity (key);
}



static void
afs_string_to_key(str, key, cell)
char		*str;
des_cblock	key;
char		*cell;                  /* cell for password */
{
    if (strlen(str) > 8) {
	afs_transarc_StringToKey (str, cell, key);
    } else {
	afs_cmu_StringToKey (str, cell, key);
    }
}
#endif /* KERBEROS */



/*
 * Initialize the authStruct data structure.
 */
static void
init_auth()
{
    if (!auth) {
	auth = (authStruct *) ch_malloc(sizeof(authStruct));
    } else {
	int i;
	if (auth->authTypes) {
	    for (i = 0; auth->authTypes[i].data != NULL; i++) {
		if (auth->authTypes[i].data != NULL) {
		    free(auth->authTypes[i].data);
		}
	    }
	    free(auth->authTypes);
	}
    }
    auth->ldaprc = 0;
    auth->msgid = 0;
    auth->timer = 0;
    auth->workingDlg = NULL;
    auth->authTextWidget = NULL;
    auth->parent = NULL;
    auth->ident = NULL;
    auth->dn = NULL;
    auth->dn_list = NULL;
    auth->numDNs = 0;
    auth->callback = NULL;
    auth->call_data = NULL;
    auth->authTypes = NULL;
    auth->activeAuth = NULL;
}



/*
 * Free the global authStruct
 */
static void
free_auth()
{
    if (auth != NULL) {
	int i;
	if (auth->authTypes) {
	    for (i = 0; auth->authTypes[i].data != NULL; i++) {
		if (auth->authTypes[i].data) {
		    free(auth->authTypes[i].data);
		}
	    }
	    free(auth->authTypes);
	}
	free(auth);
    }
}
