/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Support for searchbase browsing
 */

#include <Xm/SelectioB.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/Form.h>
#include <Xm/ScrolledW.h>
#include <Xm/PanedW.h>
#include <Xm/List.h>
#include <Xm/Protocols.h>
#include <Xm/AtomMgr.h>

#include <malloc.h>
#include <stdio.h>
#include <string.h>

#include <lber.h>
#include <ldap.h>
#include <disptmpl.h>

#include "dtypes.h"
#include "globals.h"
#include "util.h"
#include "dialog.h"
#include "menu.h"
#include "read.h"
#include "find.h"
#include "help.h"

#define	LEAFFILTER	"(objectClass=*)";
#define	NOLEAFFILTER	"(|(objectClass=quipuNonLeafObject)(objectClass=externalNonLeafObject))"

/* Forward references */
#ifndef NEEDPROTOS
static void	browseOpen();
static void	browseDblClick();
static void	browseItemSelected();
static void	browseDone();
static void	browseSetSB();
static void	browseDisplay();
static char	get_icon();
XtTimerCallbackProc	check_getchild();
static void	cancel_getchild();
static Widget	create_browser();
static void	update_browser();
static Widget	create_sb_popup();
static void	jump_searchbase();
static XmString	*result2items();
#else /* !NEEDPROTOS */
static void	browseOpen(Widget w, searchData *sd,
			XmSelectionBoxCallbackStruct *cbs);
static void	browseDblClick(Widget w, searchData *sd, 
			XmSelectionBoxCallbackStruct *cbs);
static void	browseItemSelected(Widget w, searchData *sd,
			XmSelectionBoxCallbackStruct *cbs);
static void	browseDone(Widget w, searchData *sd,
			XmSelectionBoxCallbackStruct *cbs);
static void	browseSetSB(Widget w, searchData *sd,
			XmSelectionBoxCallbackStruct *cbs);
static void	browseDisplay(Widget w, searchData *sd,
			XmSelectionBoxCallbackStruct *cbs);
static char	get_icon(char *objclass);
XtTimerCallbackProc	check_getchild(struct searchData *sd);
static void	cancel_getchild(Widget w, struct searchData *sd,
			XmAnyCallbackStruct *cbs);
static Widget	create_browser(Widget parent, searchData *sd,
			LDAPMessage *result);
static void	update_browser(Widget parent, searchData *sd,
			LDAPMessage *result);
static Widget	create_sb_popup(Widget w, searchData *sd);
static void	jump_searchbase(Widget w, searchData *sd,
			XmAnyCallbackStruct *cbs);
static XmString	*result2items(LDAPMessage *result, char ***dn_list,
			char ****objcllist, int *count);
#endif /* !NEEDPROTOS */

/* Setup for action area */
static ActionAreaItem	actions[] = {
    { "browseDoneButton",  browseDone, NULL },
    { "browseOpenButton",  browseOpen, NULL },
    { "browseDisplayButton",  browseDisplay, NULL },
    { "browseSetSearchbaseButton",  browseSetSB, NULL },
    { "browseHelpButton",  doHelpDlg, HLP_TOPIC_BROWSING },
};

static char *empty_val[] = { "", NULL }; 

#ifndef NEEDPROTOS
extern int strcasecmp();
#else /* !NEEDPROTOS */
extern int strcasecmp(char *, char *);
#endif /* !NEEDPROTOS */

/*
 * Starting at the current searchbase, initiate an asynchronous LDAP search,
 * returning all entries at the current level.  An XtTimeout will check
 * for completion.
 */
static void
get_children(w, sd)
Widget			w;
struct searchData	*sd;
{

    char		*filter;
    static char		*attrs[] = { "objectClass", NULL };
    static char		errtxt[1024];
    Widget		wdparent;
 
    if (sd == NULL) {
	sd = (struct searchData *) malloc(sizeof(struct searchData));
	sd->parent = w;
	sd->start_time = time(NULL);
	sd->searchbase = strdup(search_base);
	sd->newsb = NULL;
	sd->dn_list = NULL;
	sd->objcllist = NULL;
	sd->browseDlg = NULL;
	sd->wantLeaves = 1; /* XXX */
    }

    wdparent = sd->browseDlg ? sd->browseDlg : w;

    if (sd->wantLeaves) {
	filter = LEAFFILTER;
    } else {
	filter = NOLEAFFILTER;
    }

    /* Fire off a search */
    sd->msgid = ldap_search(ldp, (sd->newsb == NULL) ? sd->searchbase
	    : sd->newsb, LDAP_SCOPE_ONELEVEL,
	    filter, attrs, 0); 
    if (sd->msgid == -1) {
	sprintf(errtxt, "Error: %s ", ldap_err2string(ldp->ld_errno));
	errorDlg(wdparent, errtxt);
	return;
    }

    sd->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
	    (XtTimerCallbackProc) check_getchild, sd);
    sd->workingDlg = create_working_dialog(wdparent,
	    "Searching the X.500 Directory...",
	    "Cancel", NULL, NULL,
	    (XtCallbackProc) cancel_getchild, NULL, NULL,
	    sd, w);
}






/*
 * User clicked on the cancel button in the working dialog.  Abandon the
 * LDAP search and clean up.
 */
static void
cancel_getchild(w, sd, cbs)
Widget			w;
struct searchData	*sd;
XmAnyCallbackStruct	*cbs;
{
    ldap_abandon(ldp, sd->msgid);
    free(sd->newsb);
    if (sd->status) {
	displayLabelTimed(sd->status, "Cancelled", 3);
    }
    XtRemoveTimeOut(sd->timer);
}





/*
 * Called periodically to see if the LDAP operation has completed.
 */
XtTimerCallbackProc
check_getchild(sd)
struct searchData	*sd;
{
    int			rc;
    LDAPMessage		*result;
    static char		errtxt[1024];
    Widget		wdparent;

    wdparent = sd->browseDlg ? sd->browseDlg : sd->parent;

    /* See if the search operation has completed */
    rc = ldap_result(ldp, sd->msgid, 1, zerotime, &result);
    switch (rc) {
    case 0:
	/* search not yet complete */
	sd->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
		(XtTimerCallbackProc) check_getchild, sd);
	return;
	break;
    case -1:
	/* fatal error */
	/* XXX re-open LDAP connection */
	sd->ldaprc = ldp->ld_errno;
	ldap_msgfree(result);
	break;
    default:
	/* search is finished */
	sd->ldaprc = ldap_result2error(ldp, result, 0);

	if ((sd->ldaprc == LDAP_SUCCESS)
	       || (sd->ldaprc == LDAP_TIMELIMIT_EXCEEDED)
	       || (sd->ldaprc == LDAP_SIZELIMIT_EXCEEDED)) {
	    /* XXX notify when limits exceeded */
	    /* search ok, display results */
	    if (sd->newsb) {
		free(sd->searchbase);
		sd->searchbase = strdup(sd->newsb);
	    }

	    /* If there's a working dialog on-screen, destroy it */
	    if (sd->workingDlg != NULL) {
		XtDestroyWidget(sd->workingDlg);
		sd->workingDlg = NULL;
	    }

	    /* Sort the entries */
	    if (ldap_sort_entries(ldp, &result, NULL, strcasecmp) != 0) {
		errorDlg(sd->parent, "Error sorting entries");
		return;
	    }

	    /* Check to see if we need to create browse dialog */
	    if (sd->browseDlg == NULL) {
		/* No browse dlg yet - create one */
		sd->browseDlg = create_browser(sd->parent, sd, result);
		XtRealizeWidget(sd->browseDlg);
		XtMapWidget(sd->browseDlg);
		return;
	    } else {  /* already have a browser */
		update_browser(sd->parent, sd, result);
		return;
	    }
	}
    }
    /* error */
    sprintf(errtxt, "Error while searching: %s", ldap_err2string(sd->ldaprc));
    errorDlg(wdparent, errtxt);
    /* If there's a working dialog on-screen, destroy it */
    if (sd->workingDlg != NULL) {
	XtDestroyWidget(sd->workingDlg);
	sd->workingDlg = NULL;
    }
    return;
}




		
/*
 * Given an object class, return the appropriate char to use
 * for the icon (which appears to the left of the object's cn
 * in the browser.  Note - not implemented in 1.0b1, mostly
 * because I'm a crummy artist.  Really, icons next to items
 * in the browser should be done by making a browse widget,
 * instead of defining a font with icons for characters.
 */
static char
get_icon(objclass)
char *objclass;
{
#ifdef notyet
    /*
     * We bagged doing icons in the browser for now, until we find
     * someone with some shred of artistic talent.
     */
    int i;

    for (i = 0; objClasses[i].name != NULL; i++) {
	if (!strcmp(objclass, objClasses[i].name)) {
	    return(objClasses[i].icon);
	}
    }
    return('g');
#endif
    return((char) 0x0c);
}
	




/*
 * Callback routine for browse buttons and browse menu choices
 */
void
do_browse(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    if (browserActive) {
	/*
	 * If we're in the process of constructing the browse
	 * dialog, browseDlg might still be NULL.
	 */
	if (browseDlg != NULL) {
	    XtMapWidget(browseDlg);
	    XtPopup(browseDlg, XtGrabNone);
	}
    } else {
	browserActive = 1;
	get_children(toplevel, NULL);
    }
}




/*
 * Called when an LDAP operation completes, and we have a new list
 * of objects to display.
 */
static void
update_browser(w, sd, result)
Widget		w;
searchData	*sd;
LDAPMessage	*result;
{
    XmString	*items;
    int		count;
    Widget	parent;


    XtRemoveCallback(sd->listWidget, XmNdefaultActionCallback,
	    (XtCallbackProc) browseDblClick, sd);

    items = result2items(result, &(sd->dn_list), &(sd->objcllist), &count);
    ldap_msgfree(result);
    XmListDeleteAllItems(sd->listWidget);
    XmListAddItems(sd->listWidget, items, count, 1);
    XmListSelectPos(sd->listWidget, 1, False);
    freeXmStringList(items, count);
    parent = XtParent(sd->browseOption);
    XtDestroyWidget(sd->browseOption);
    sd->browseOption = create_sb_popup(parent, sd);
    XtManageChild(sd->browseOption);
    XtAddCallback(sd->listWidget, XmNdefaultActionCallback,
	    (XtCallbackProc) browseDblClick, sd);
}




/*
 * Convert an LDAP result message to an array of XmStrings.  Also return a
 * null-terminated array of DNs and a null-terminated array of null-terminated
 * arrays of object classes for each item (used by routines which examine
 * the object class list for an entry to try to figure out if it's a leaf
 * node.
 *
 * XXX - the method for determining if an object is a leaf object is,
 * well, not really good.  It's as good as we can do, given that we
 * can't find out except by attempting to do a onelevel search for 
 * anything and see if we get any results.  Since that would be so
 * expensive, we just look at the objectClass of the entry and take
 * a guess.
 */ 
static XmString
*result2items(result, dn_list, objcllist, count)
LDAPMessage	*result;
char		***dn_list;
char		****objcllist;
int		*count;
{
    XmString	*items, s1, s2;
    LDAPMessage	*ep;
    int		i, j;
    char	**xdn, *fr_name;
    char	**vals;
    char	item[512];
    char	iconstr[2];

    *count = ldap_count_entries(ldp, result);

    /* free any existing lists */
    if (dn_list && *dn_list) {
	for (i = 0; (*dn_list)[i] != NULL; i++) {
	    free((*dn_list)[i]);
	}
	free(*dn_list);
    }

    if (objcllist && *objcllist) {
	for (i = 0; (*objcllist)[i] != NULL; i++) {
	    ldap_value_free((*objcllist)[i]);
	}
	free(*objcllist);
    }

    if (!(*count)) {
	*dn_list = NULL;
	*objcllist = NULL;
	return NULL;
    }

    *dn_list = (char **) malloc ((*count + 1) * sizeof(char *));
    *objcllist = (char ***) malloc ((*count + 1) * sizeof(char **));
    items = (XmStringTable) XtMalloc((*count + 1) * sizeof(XmString));

    for (ep = ldap_first_entry(ldp, result), i = 0; ep != NULL;
	    ep = ldap_next_entry(ldp, ep), i++) {
	(*dn_list)[i] = ldap_get_dn(ldp, ep);
	xdn = ldap_explode_dn((*dn_list)[i], TRUE);
	vals = ldap_get_values(ldp, ep, "objectClass");
	/* vals may be NULL (damn Swiss!) - careful */
	if (vals == NULL) {
	    (*objcllist)[i] = copy_vals(empty_val);
	} else {
	    (*objcllist)[i] = vals;
	}
	j = 0;
	while ((vals != NULL) && vals[j] != NULL) {
	    if (!strcasecmp(vals[j], "top")) {
		j++;
		continue;
	    }
	    if (!strcasecmp(vals[j], "quipuObject")) {
		j++;
		continue;
	    }
	    if (!strcasecmp(vals[j], "quipuNonLeafObject")) {
		j++;
		continue;
	    }
	    if (!strcasecmp(vals[j], "xaclObject")) {
		j++;
		continue;
	    }
	    if (!strcasecmp(vals[j], "pilotObject")) {
		j++;
		continue;
	    }
	    break;
	}
	/* Convert to friendly name */
	fr_name = ldap_friendly_name(appConfig.friendlyMap, xdn[0], &fr_map);
	if (fr_name == NULL) {
	    fr_name = xdn[0];
	}
	sprintf(item, "%s ", fr_name);
#ifdef notyet /* No icon hints yet */
	sprintf(iconstr, "%c", get_icon(vals[j]));
	s1 = XmStringCreate(iconstr, "iconfont");
	s2 = XmStringCreate(item, "helv");
	s3 = XmStringCreate(vals ? vals[j] : "unknown object type");
	(items)[i] = XmStringConcat(s1, s2);
#else
	s1 = XmStringCreate(item, "browserfont");
	sprintf(item, "[%s]", vals ? vals[j] : "unknown object type");
	s2 = XmStringCreate(item, "browserfont2");
	(items)[i] = XmStringConcat(s1, s2);
	XmStringFree(s1);
	XmStringFree(s2);
#endif
	(void) ldap_value_free(xdn);
#ifdef notyet
	XmStringFree(s1);
	XmStringFree(s2);
#endif
    }
    (*dn_list)[i] = NULL;
    (*objcllist)[i] = NULL;
    return items;
}





/*
 * Set up the browse dialog
 */
static Widget
create_browser(parent, sd, result)
Widget		parent;
searchData	*sd;
LDAPMessage	*result;
{
    Widget	browseList, browsePane, browseAction, browseCntrlForm,
		findButton;
    XmString	*items;
    XmString	s;
    int		count;
    int		i;
    Arg		args[10];
    Atom	wm_delete_window;


    browseDlg = XtVaAppCreateShell("browseDlg", "Xax500",
	    applicationShellWidgetClass,
	    XtDisplay(toplevel),
	    XmNtitle, "Browser",
	    NULL);
    sd->browseDlg = browseDlg;
    wm_delete_window = XmInternAtom(XtDisplay(browseDlg),
	    "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(browseDlg, wm_delete_window, 
	    (XtCallbackProc) browseDone, sd);

    /* Paned window holds all the other widgets */
    browsePane = XtVaCreateWidget("browsePane",
	    xmPanedWindowWidgetClass,
	    browseDlg,
	    XmNsashWidth, 1,
	    XmNsashHeight, 1,
	    NULL);

    /* Status area */
    s = XmStringCreateLocalized(" ");
    i = 0;
    XtSetArg(args[i], XmNlabelString, s); i++;
    XtSetArg(args[i], XmNleftAttachment, XmATTACH_FORM); i++;
    XtSetArg(args[i], XmNrightAttachment, XmATTACH_FORM); i++;
    XtSetArg(args[i], XmNtopAttachment, XmATTACH_FORM); i++;
    sd->status = XtCreateManagedWidget("FindStatus",
	    xmLabelWidgetClass,
	    browsePane, args, i);
	XmStringFree(s);

    /*
     * A small form widget to hold the various controls for this
     * window.
     */
    browseCntrlForm =  XtVaCreateWidget("browseCntrlForm",
	    xmFormWidgetClass, browsePane,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    NULL);

    /* Create a find button - shortcut to raise find window */
    findButton = create_pixmap_button(browseCntrlForm, "find_image",
	    NULL, "findButton");
    XtVaSetValues(findButton,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNtopAttachment, XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    NULL);

    XtAddCallback(findButton, XmNactivateCallback,
	    (XtCallbackProc) raiseFindWindow, NULL);
    XtManageChild(findButton);
    XtManageChild(browseCntrlForm);

    items = result2items(result, &(sd->dn_list), &(sd->objcllist), &count);

    /* Create a scrolled list */
    i = 0;
    XtSetArg(args[i], XmNlistSizePolicy, XmRESIZE_NONE); i++;
    browseList = XmCreateScrolledList(browsePane, "browseList", args, i);
    XtVaSetValues(browseList,
	    XmNitems, items,
	    XmNitemCount, count,
	    XmNvisibleItemCount, count,
	    NULL);
    sd->listWidget = browseList;
    XtAddCallback(browseList, XmNdefaultActionCallback,
	    (XtCallbackProc) browseDblClick, sd);
    XtAddCallback(browseList, XmNbrowseSelectionCallback,
	    (XtCallbackProc) browseItemSelected, sd);
    XtManageChild(browseList);

    sd->browseOption = create_sb_popup(browseCntrlForm, sd);

    /* Create an action area */
    actions[0].data = (caddr_t) sd;
    actions[1].data = (caddr_t) sd;
    actions[2].data = (caddr_t) sd;
    actions[3].data = (caddr_t) sd;
    browseAction = CreateActionArea(browsePane, actions, XtNumber(actions));

    XtManageChild(sd->browseOption);
    XmListSelectPos(sd->listWidget, 1, False);

    for (i = 0; i < count; i++) {
	XmStringFree(items[i]);
    }

    ldap_msgfree(result);
    XtManageChild(browsePane);
    return(browseDlg);
}



/*
 * Attempt to set the searchbase and get the children of the selected item.
 */
static void
browseOpen(w, sd, cbs)
Widget				w;
searchData			*sd;
XmSelectionBoxCallbackStruct	*cbs;
{
    char	*dn;
    int		*pos_list, pos_count;

    if (XmListGetSelectedPos(sd->listWidget, &pos_list, &pos_count) == TRUE) {
	dn = sd->dn_list[*pos_list - 1];
	sd->newsb = strdup(dn);
    } else {
	return;
    }
		
    get_children(w, sd);
}




/*
 * Handle a double-click in the browser.  Make an intelligent guess about
 * the selected item (e.g. if it's a leaf node or not) and do the
 * appropriate thing (set searchbase, or read and display the entry).
 */
static void
browseDblClick(w, sd, cbs)
Widget				w;
searchData			*sd;
XmSelectionBoxCallbackStruct	*cbs;
{
    char	*dn;
    int		*pos_list, pos_count;
    char	**p;

    if (XmListGetSelectedPos(sd->listWidget, &pos_list, &pos_count) == TRUE) {
	dn = sd->dn_list[*pos_list - 1];
	/* See what this thing is */
	p = sd->objcllist[*pos_list - 1];
	if (is_leaf(dn, p)) {
	    do_read(w, ldp, dn, sd->objcllist[*pos_list - 1], sd->status);
	} else {
	    sd->newsb = strdup(dn);
	    get_children(w, sd);
	}
    } else {
	return;
    }
}

	


static void
browseItemSelected(w, sd, cbs)
Widget				w;
searchData			*sd;
XmSelectionBoxCallbackStruct	*cbs;
{
}




/*
 * Get rid of browser.
 */
static void
browseDone(w, sd, cbs)
Widget				w;
searchData			*sd;
XmSelectionBoxCallbackStruct	*cbs;
{
    int	i;

    browserActive = 0;
    browseDlg = NULL;

    /* clean up */
    XtDestroyWidget(sd->browseDlg);
    /* free any existing lists */
    if (sd->dn_list) {
	for (i = 0; sd->dn_list[i] != NULL; i++) {
	    free(sd->dn_list[i]);
	}
	free(sd->dn_list);
    }

    if (sd->objcllist) {
	for (i = 0; sd->objcllist[i] != NULL; i++) {
	    ldap_value_free(sd->objcllist[i]);
	}
	free(sd->objcllist);
    }
    if (sd->searchbase) {
	free(sd->searchbase);
    }
    free(sd);
}



/*
 * Change the global searchbase.  Updates the find window's display.
 */
static void
browseSetSB(w, sd, cbs)
Widget				w;
searchData			*sd;
XmSelectionBoxCallbackStruct	*cbs;
{
    if (sd->searchbase) {
	strcpy(search_base, sd->searchbase);
	updateSearchBaseText(search_base);
    }
}



/*
 * Display the selected item.
 */
static void
browseDisplay(w, sd, cbs)
Widget				w;
searchData			*sd;
XmSelectionBoxCallbackStruct	*cbs;
{
    char	*dn;
    int		*pos_list, pos_count;

    if (XmListGetSelectedPos(sd->listWidget, &pos_list, &pos_count) == TRUE) {
	dn = sd->dn_list[*pos_list - 1];
	do_read(w, ldp, dn, sd->objcllist[*pos_list - 1], sd->status);
    } else {
	return;
    }
}



/*
 * Handle a selection in the option menu (allows one to move back up the
 * directory tree towards the root.
 */
static void
jump_searchbase(w, sd, cbs)
Widget			w;
searchData		*sd;
XmAnyCallbackStruct	*cbs;
{
    Widget	rc, button;
    WidgetList	children;
    int		i, num_children, selected, ssize;
    char	**bases;

    /* Figure out which component of the searchbase was selected */
    rc = XtParent(w);
    XtVaGetValues(rc, XmNmenuHistory, &button, NULL);
    XtVaGetValues(rc,
	    XmNnumChildren, &num_children,
	    XmNchildren, &children,
	    NULL);
    for (i = 0; i < num_children; i++) {
	if (children[i] == button) {
	    selected = i;
	    break;
	}
    }

    /* Construct the new destination searchbase */
    bases = ldap_explode_dn(sd->searchbase, 0);
    for (ssize = 0, i = (num_children - selected - 1); i < (num_children - 1);
	    i++) {
	ssize += (strlen(bases[i]) + 2);
    }
    ssize++;
    sd->newsb = (char *) malloc(ssize);
    *(sd->newsb) = '\0';
    for (i = (num_children - selected - 1); i < (num_children - 1); i++) {
	strcat(sd->newsb, bases[i]);
	strcat(sd->newsb, ", ");
    }
    sd->newsb[strlen(sd->newsb) - 2] = '\0';
    ldap_value_free(bases);

    /* Go there and display children */
    get_children(sd->browseDlg, sd);
}




/*
 * Create the option menu which allows the user to navigate back toward the
 * root of the directory.
 */
static Widget
create_sb_popup(w, sd)
Widget		w;
searchData	*sd;
{
    Widget	option;
    char	**bases;
    char	**bases_notypes;
    MenuItem	*bases_items;
    int		num_components, i, j;
	
    /*
     * Create an option popup with the path from the current searchbase
     * back to the root.
     */
    bases = ldap_explode_dn(sd->searchbase, 0);
    bases_notypes  = ldap_explode_dn(sd->searchbase, 1);

    /* How deep are we? */
    num_components = 0;

    /* If we're at root, bases_notypes[0] will be a zero-length string */
    while ((bases[num_components] != NULL) &&
	strlen(bases_notypes[num_components])) {
	num_components++;
    }
    num_components++;	/* make room for root */

    bases_items = (MenuItem *) malloc((num_components + 1) * sizeof(MenuItem));
    /* root is a special case */
    bases_items[0].label = NULL;
    bases_items[0].widgetname = "browserWorldButton";
    bases_items[0].class = &xmPushButtonGadgetClass;
    bases_items[0].mnemonic = '\0';
    bases_items[0].accelerator = "";
    bases_items[0].accel_text = "";
    bases_items[0].callback = jump_searchbase;
    bases_items[0].callback_data = (XtPointer) sd;
    bases_items[0].subitems = NULL;
    bases_items[0].ret_widget = NULL;
    for (i = 1, j = (num_components - 2); i < num_components; i++, j--) {
	bases_items[i].label = bases_notypes[j];
	bases_items[i].widgetname = bases_notypes[j];
	bases_items[i].class = &xmPushButtonGadgetClass;
	bases_items[i].mnemonic = '\0';
	bases_items[i].accelerator = "";
	bases_items[i].accel_text = "";
	bases_items[i].callback = jump_searchbase;
	bases_items[i].callback_data = (XtPointer) sd;
	bases_items[i].subitems = NULL;
	bases_items[i].ret_widget = NULL;
    }
    bases_items[num_components].widgetname = NULL;
    option = BuildMenu(w, XmMENU_OPTION, NULL, '\0',
	    bases_items, num_components - 1);
    XtVaSetValues(option,
	    XmNtopAttachment, XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    XmNleftAttachment, XmATTACH_FORM,
	    NULL);
    ldap_value_free(bases);
    ldap_value_free(bases_notypes);
    free(bases_items);
    return(option);
}
