/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <Xm/RowColumn.h>
#include <Xm/CascadeBG.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/LabelG.h>
#include <Xm/SeparatoG.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <stdio.h>

#include <lber.h>
#include <ldap.h>

#include "dtypes.h"
#include "globals.h"
#include "menu.h"
#include "about.h"
#include "find.h"
#include "prefs.h"
#include "util.h"
#include "auth.h"
#include "browse.h"
#include "dialog.h"
#include "help.h"
#include "newentry.h"

#ifdef NEEDPROTOS
void junk_cb();
#else /* !NEEDPROTOS */
void junk_cb(Widget w, int item_no, XmAnyCallbackStruct *cbs);
#endif /* !NEEDPROTOS */

/* 
 * note - in Motif 1.2.2, use of XmVaSEPARATOR in arg list will cause
 * button names to get lost.
 */

Widget	menuitem_NewEntry;
Widget	menuitem_PageSetup;
Widget	menuitem_Print;
Widget	menuitem_Preferences;
Widget	menuitem_RetrieveNews;
Widget	menuitem_Cut;
Widget	menuitem_Copy;
Widget	menuitem_Paste;
Widget	menuitem_Clear;
Widget	menuitem_CopyEntryName;
Widget	menuitem_Authenticate;
Widget	menuitem_DiscardAuthentication;
Widget	menuitem_FlushCache;
Widget	menuitem_ModifyThisMenu;
Widget	menuitem_ShowStatus;
Widget	menuitem_RecentlyUsed;
Widget	menuitem_Help;

/* Build popup, option and pulldown menus, depending on the menu_type.
 * It may be XmMENU_PULLDOWN, XmMENU_OPTION or  XmMENU_POPUP.  Pulldowns
 * return the CascadeButton that pops up the menu.  Popups return the menu.
 * Option menus are created, but the RowColumn that acts as the option
 * "area" is returned unmanaged. (The user must manage it.)
 * Pulldown menus are built from cascade buttons, so this function
 * also builds pullright menus.  The function also adds the right
 * callback for PushButton or ToggleButton menu items.
 */
Widget
BuildMenu(parent, menu_type, menu_title, menu_mnemonic, items, sel_item)
Widget parent;
int menu_type;
char *menu_title;
char menu_mnemonic;
MenuItem *items;
int sel_item;
{
    Widget menu, cascade, widget;
    int i;
    XmString str;

    if (menu_type == XmMENU_PULLDOWN || menu_type == XmMENU_OPTION)
        menu = XmCreatePulldownMenu(parent, "_pulldown", NULL, 0);
    else if (menu_type == XmMENU_POPUP)
        menu = XmCreatePopupMenu(parent, "_popup", NULL, 0);
    else {
        XtWarning("Invalid menu type passed to BuildMenu()");
        return NULL;
    }

    /* Pulldown menus require a cascade button to be made */
    if (menu_type == XmMENU_PULLDOWN) {
        str = XmStringCreateSimple(menu_title);
        cascade = XtVaCreateManagedWidget(menu_title,
            xmCascadeButtonGadgetClass, parent,
            XmNsubMenuId,   menu,
            /* XmNlabelString, str, */
            XmNmnemonic,    menu_mnemonic,
            NULL);
        XmStringFree(str);
    } else if (menu_type == XmMENU_OPTION) {
        /* Option menus are a special case, but not hard to handle */
        Arg args[2];
        str = XmStringCreateSimple(menu_title);
        XtSetArg(args[0], XmNsubMenuId, menu);
        /* XtSetArg(args[1], XmNlabelString, str); */
        /* This really isn't a cascade, but this is the widget handle
         * we're going to return at the end of the function.
         */
        cascade = XmCreateOptionMenu(parent, menu_title, args, 1);
        XmStringFree(str);
    }

    /* Now add the menu items */
    for (i = 0; (items != NULL) && (items[i].widgetname != NULL); i++) {
	XmString tmp;
        /* If subitems exist, create the pull-right menu by calling this
         * function recursively.  Since the function returns a cascade
         * button, the widget returned is used..
         */
        if (items[i].subitems)
            if (menu_type == XmMENU_OPTION) {
                XtWarning("You can't have submenus from option menu items.");
                continue;
            } else
                widget = BuildMenu(menu, XmMENU_PULLDOWN,
                    items[i].label, items[i].mnemonic, items[i].subitems, 0);
        else
            widget = XtVaCreateManagedWidget(items[i].widgetname,
					     *items[i].class, menu,
					     NULL);

	    if (items[i].label != NULL) {
	        tmp = XmStringCreateSimple(items[i].label);
		XtVaSetValues(widget, XmNlabelString, tmp, NULL);
		XmStringFree(tmp);
	    }

	    if (items[i].ret_widget != NULL) {
		*items[i].ret_widget = widget;
	    }

	    /* If option menu, set default selected item */
            if ((menu_type == XmMENU_OPTION) && (i == sel_item)) {
		XtVaSetValues(menu, XmNmenuHistory, widget, NULL);
	    }

        /* Whether the item is a real item or a cascade button with a
         * menu, it can still have a mnemonic.
         */
        if (items[i].mnemonic)
            XtVaSetValues(widget, XmNmnemonic, items[i].mnemonic, NULL);

        /* any item can have an accelerator, except cascade menus. But,
         * we don't worry about that; we know better in our declarations.
         */
        if (items[i].accelerator) {
            str = XmStringCreateSimple(items[i].accel_text);
            XtVaSetValues(widget,
                XmNaccelerator, items[i].accelerator,
                XmNacceleratorText, str,
                NULL);
            XmStringFree(str);
        }

        if (items[i].callback)
            XtAddCallback(widget,
                (items[i].class == &xmToggleButtonWidgetClass ||
                 items[i].class == &xmToggleButtonGadgetClass)?
                    XmNvalueChangedCallback : /* ToggleButton class */
                    XmNactivateCallback,      /* PushButton class */
                items[i].callback, items[i].callback_data);
    }

    /* for popup menus, just return the menu; pulldown menus, return
     * the cascade button; option menus, return the thing returned
     * from XmCreateOptionMenu().  This isn't a menu, or a cascade button!
     */
    return menu_type == XmMENU_POPUP? menu : cascade;
}


MenuItem file_items[] = {
  { "NewEntryButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, doNewEntry, NULL, NULL, &menuitem_NewEntry },
  {"_sep1", "_sep1", &xmSeparatorGadgetClass,
   '\0', NULL, NULL, NULL, NULL, NULL, NULL },
  { "PrintButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_Print },
  {"_sep1", "_sep1", &xmSeparatorGadgetClass,
   '\0', NULL, NULL, NULL, NULL, NULL, NULL },
  { "PreferencesButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_Preferences },
  {"_sep1", "_sep1", &xmSeparatorGadgetClass,
   '\0', NULL, NULL, NULL, NULL, NULL, NULL },
  { "ShowStatusButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_ShowStatus },
  { "RetrieveNewsButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, retrieveNews, NULL, NULL, &menuitem_RetrieveNews},
  {"_sep1", "_sep1", &xmSeparatorGadgetClass,
   (char) NULL, NULL, NULL, NULL, NULL, NULL, NULL },
  { "QuitButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, cleanup_and_exit, NULL, NULL, NULL},

  {"_sep1", "_sep1", &xmSeparatorGadgetClass,
   '\0', NULL, NULL, NULL, NULL, NULL, NULL },
  NULL
};
	

MenuItem edit_items[] = {
  {"_sep1", "_sep1", &xmSeparatorGadgetClass,
   '\0', NULL, NULL, NULL, NULL, NULL, NULL },
  { "CutButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_Cut },
  { "CopyButton", NULL, &xmPushButtonGadgetClass,
  '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_Copy },
  { "PasteButton", NULL , &xmPushButtonGadgetClass,
    '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_Paste },
  { "ClearButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_Clear },
  {"_sep1", "_sep1", &xmSeparatorGadgetClass,
   '\0', NULL, NULL, NULL, NULL, NULL, NULL },
  {"_sep1", "_sep1", &xmSeparatorGadgetClass,
   '\0', NULL, NULL, NULL, NULL, NULL, NULL },
  { "CopyEntryNameButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_CopyEntryName },
  NULL
};

MenuItem directory_items[] = {
  { "AuthenticateButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, do_auth, NULL, NULL, &menuitem_Authenticate },
  { "DiscardAuthenticationButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, do_authDiscard, NULL, NULL,
    &menuitem_DiscardAuthentication },
  { "FlushCacheButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, flushCache, NULL, NULL, &menuitem_FlushCache },
  NULL
};

MenuItem searchbase_items[] = {
  { "BrowseButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, do_browse, NULL, NULL, NULL },
  { "RecentlyUsedButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_RecentlyUsed },
  { "ModifyThisMenuButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, junk_cb, NULL, NULL, &menuitem_ModifyThisMenu },
  {"_sep1", "_sep1", &xmSeparatorGadgetClass,
   '\0', NULL, NULL, NULL, NULL, NULL, NULL },
  NULL
};


MenuItem window_items[] = {
  NULL
};

MenuItem help_items[] = {
  { "AboutButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, doAboutDlg, NULL, NULL, NULL },
  { "HelpButton", NULL, &xmPushButtonGadgetClass,
    '\0', NULL, NULL, doHelpDlg, "Main Index", NULL, &menuitem_Help },
  NULL
};

void
create_menus(parent, MenuBar)
Widget	*parent;
Widget	*MenuBar;
{
    Widget		FileMenu, EditMenu, DirectoryMenu, SearchbaseMenu,
			HelpMenu;

    /* Set up menu bar */
    *MenuBar = XmCreateMenuBar(*parent, "MenuBar", NULL, 0);

    /* Set up the file menu */
    FileMenu = BuildMenu(*MenuBar, XmMENU_PULLDOWN, "File", 'F', file_items, 0);
    EditMenu = BuildMenu(*MenuBar, XmMENU_PULLDOWN, "Edit", 'E', edit_items, 0);
    DirectoryMenu = BuildMenu(*MenuBar, XmMENU_PULLDOWN, "Directory", 'D',
	    directory_items, 0);
    SearchbaseMenu = BuildMenu(*MenuBar, XmMENU_PULLDOWN, "Searchbase", 'S',
	    searchbase_items, 0);
    HelpMenu = BuildMenu(*MenuBar, XmMENU_PULLDOWN, "Help", 'H', help_items, 0);
	
    /* Grey out unimplemented menus */
    XtSetSensitive(menuitem_Print, False);
    XtSetSensitive(menuitem_Preferences, False);
    XtSetSensitive(menuitem_Cut, False);
    XtSetSensitive(menuitem_Copy, False);
    XtSetSensitive(menuitem_Paste, False);
    XtSetSensitive(menuitem_Clear, False);
    XtSetSensitive(menuitem_CopyEntryName, False);
    XtSetSensitive(menuitem_ModifyThisMenu, False);
    XtSetSensitive(menuitem_ShowStatus, False);
    XtSetSensitive(menuitem_RecentlyUsed, False);
    /* XtSetSensitive(menuitem_Help, False); */

    XtManageChild(*MenuBar); 
}



void
junk_cb(w, item_no, cbs)
Widget			w;
int			item_no;
XmAnyCallbackStruct	*cbs;
{
    printf("No callback for this action yet\n");
}
