/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Read single objects.
 */

#include <Xm/SelectioB.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/MessageB.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/ScrolledW.h>

#include <lber.h>
#include <ldap.h>
#include <disptmpl.h>

#include <stdio.h>
#include <sys/time.h>

#include "dtypes.h"
#include "util.h"
#include "menu.h"
#include "globals.h"
#include "read.h"
#include "display.h"
#include "bitmaps.h"
#include "dialog.h"


/* forward declarations */
#ifndef NEEDPROTOS
void	check_read();
void	cancel_read();
Boolean	entry_on_screen();
void	raise_entry();
#else /* !NEEDPROTOS */
void	check_read(struct searchData *f);
void	cancel_read(Widget w, XtPointer client_data, XmAnyCallbackStruct *cbs);
Boolean	entry_on_screen(char *dn);
void	raise_entry(char *dn);
#endif /* !NEEDPROTOS */

/*
 * Retrieve the entry named by "dn".  w is the parent widget.  ld is the
 * LDAP struct (need to be initialized), dn is the distinguished name to
 * read, ocl is an array of object classes for the object, and status
 * should point to a text widget into which status information will be
 * put.
 *
 * ocl can be NULL, in which case do_read() will perform two reads - one
 * to get the object class, and one to retrieve the appropriate attributes
 * after it's decided what kind of entry it's retrieving.
 *
 * status can be null.
 */
void
do_read(w, ld, dn, ocl, status)
Widget		w;
LDAP		*ld;
char		*dn;
char		**ocl;
Widget		status;
{
    struct searchData	*f;
    char		tmp[1024];
    static char		*ocattrs[] = {"objectClass", NULL};

    /* "other, class, kerberos, auth" attributes */
    static char		*ockaattrs[] = {"objectClass", "krbName", NULL, NULL};
    char		**nattrs;

    /* Is this entry already on-screen? */
    if (entry_on_screen(dn)) {
	raise_entry(dn);
	return;
    }

    /* set up our per-search struct */
    if ((f = get_searchdata()) == NULL) {
	fprintf(stderr, "out of memory in do_read\n");
	exit(1);
    }
    f->searchbase = strdup(search_base);
    f->workingDlg = (Widget) NULL;
    if (w == NULL) {
	f->parent = toplevel;
    } else {
	f->parent = w;
    }
    f->status = status;
    f->start_time = time(NULL);
    f->etype = NULL;
    f->search_string = strdup(dn);
    f->objcllist = NULL;

    if (ocl == NULL) {
	/* retrieve the objectClass attribute */
	f->msgid = ldap_search(ld, dn, LDAP_SCOPE_BASE, "objectclass=*",
		ocattrs, 0);
	/* Pop up the working dialog */
	f->workingDlg = create_working_dialog(f->parent, "Examining entry...",
		"Cancel", NULL, NULL,
		(XtCallbackProc) cancel_read, NULL,
		NULL, f, f->parent);
    } else {
	f->etype = ldap_oc2template(ocl, templates);
	if (f->etype == NULL) {
	    sprintf(tmp, "Unknown obj type - cannot");
	    strcat(tmp, " display (until new xax500");
	    strcat(tmp, " version)");
	    errorDlg(f->parent, tmp);
	    free_searchdata(&f);
	    return;
	}

	/* Figure out which attributes we want to retrieve */
#ifdef notyet
	nattrs = ldap_tmplattrs(f->etype, ocattrs, 1, LDAP_SYN_OPT_DEFER);
#endif
	ockaattrs[2] = f->etype->dt_authattrname;	/* XXX */
	nattrs = ldap_tmplattrs(f->etype, ockaattrs, 0, 0L);

	/* fire off the search (retrieval) */
	f->msgid = ldap_search(ld, dn, LDAP_SCOPE_BASE, "objectclass=*",
		nattrs, 0);
	ldap_value_free(nattrs);

	/* Pop up the working dialog */
	f->workingDlg = create_working_dialog(f->parent, "Reading...",
		"Cancel", NULL, NULL,
		(XtCallbackProc) cancel_read, NULL,
		NULL, f, f->parent);

    }

    /* Set up an Xt App timeout function to handle the result */
    f->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
	    (XtTimerCallbackProc) check_read, f);

}






/*
 * Cancel a pending read.
 */
void
cancel_read(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    struct searchData	*f;
    XmString		s;

    f = (searchData *) client_data;

    /* abandon the current LDAP operation */
    ldap_abandon(ldp, f->msgid);

    if (f->status) {
	s = XmStringCreateLtoR("Read cancelled", XmSTRING_DEFAULT_CHARSET);
	XtVaSetValues(f->status, XmNlabelString, s, NULL);
	XmStringFree(s);
    }
	
    XtRemoveTimeOut(f->timer);

    /* Free the searchData struct */
    free(f);
}




/*
 * Check to see if the pending LDAP read operation is finished.  If
 * not, install another XtAppTimeOut and exit.  If the search is
 * complete, then either bring up an entry window or a scrolling list
 * of matches (if there was more than one match).
 */
void
check_read(f)
struct searchData *f;
{
    int			rc, matches;
    LDAPMessage		*result, *e;
    char		errtxt[512];
    char		*dn = NULL;
    char		**ocl = NULL;
    entryData		*en;

    /* See if the search operation has completed */
    rc =  ldap_result(ldp, f->msgid, 1, zerotime, &result);
    switch (rc) {
    case 0:
	/* search not yet complete */
	if ((time(NULL) - f->start_time) >
	    appConfig.ldapReadTimeout) {
	    errorDlg(f->parent, "Read operation timed out");
	    if (result) {
		ldap_msgfree(result);
	    }
	    destroyWorking(&(f->workingDlg));
	    ldap_abandon(ldp, f->msgid);
	    /* XXX clean up */
	    return;
	}
	f->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
		(XtTimerCallbackProc) check_read, f);
	return;
	break;
    case -1:
	/* fatal error */
	f->ldaprc = ldp->ld_errno;
	ldap_msgfree(result);
	if ((reconnect_ldap(f->ldaprc)) == 0) {
	    /* reconnect successful, retry */
	    do_read(f->parent, ldp, f->search_string, NULL, f->status);
	    return;
	}
	break;
    default:
	/* read is finished */
	/* If there's a working dialog on-screen, destroy it */
	    destroyWorking(&(f->workingDlg));
	if (f->status) {
	    displayLabelTimed(f->status, "Read completed", 3);
	}
    
	f->ldaprc = ldap_result2error(ldp, result, 0);
	if ((f->ldaprc == LDAP_SUCCESS) 
	       || (f->ldaprc == LDAP_SIZELIMIT_EXCEEDED)
	       || (f->ldaprc == LDAP_SIZELIMIT_EXCEEDED)) {
	    /* search ok, display results */
	    matches = ldap_count_entries(ldp, result);
	    if (matches == 1) {
		e = ldap_first_entry(ldp, result);
		if (f->etype) {
		    /* Display the one match */
		    en = ldapMessage2entryData(ldp, e, f->etype);
		    display_entry(en, f->etype);
		} else {
		    ocl = ldap_get_values(ldp, e, "objectClass");
		    dn = ldap_get_dn(ldp, e);
		    f->etype = ldap_oc2template(ocl, templates);
		    do_read(f->parent, ldp, dn, ocl, f->status);
		    free(dn);
		    ldap_value_free(ocl);
		}
		ldap_msgfree(result);
		free_searchdata(&f);
		return;
	    } else {
		/*
		 * This is bad - do_read should never result
		 * in more than one match.
		 */
		fprintf(stderr, "do_read got multiple match\n");
		return;
	    }
	}
	break;
    } /* end case */

    /* some kind of an error occurred */
    sprintf(errtxt, "Error while searching: %s", ldap_err2string(f->ldaprc));
    errorDlg(f->parent, errtxt);
    return;
}



Boolean
entry_on_screen(dn)
char *dn;
{
    displayData	*d;

    for (d = displayDlgData; d != NULL; d = d->di_next) {
	if (!strcasecmp(dn, d->di_entryData->en_dn)) {
	    return(1);
	}
    }
    return(0);
}


void
raise_entry(dn)
char *dn;
{
    displayData	*d;

    for (d = displayDlgData; d != NULL; d = d->di_next) {
	if (!strcasecmp(dn, d->di_entryData->en_dn)) {
	    XtMapWidget(d->di_displayDlg);
	    XtPopup(d->di_displayDlg, XtGrabNone);
	    latestDisplayDlg = d;
	}
    }
}
