/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Support for searching
 */

#include <Xm/SelectioB.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/MessageB.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/ScrolledW.h>

#include <lber.h>
#include <ldap.h>
#include <srchpref.h>

#include <stdio.h>
#include <sys/time.h>

#include "dtypes.h"
#include "util.h"
#include "menu.h"
#include "globals.h"
#include "search.h"
#include "read.h"
#include "select.h"
#include "bitmaps.h"
#include "dialog.h"
#include "errtxt.h"

/*
 * Cancel a pending search.
 */
void
cancel_search(findDlg, client_data, cbs)
Widget			findDlg;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    struct searchData	*f;

    /* Get the searchData struct */
    f = (searchData *) client_data;

    /* abandon the current LDAP operation */
    ldap_abandon(ldp, f->msgid);

    if (f->status) {
	updateText(f->status, "Search cancelled");
    }
	
    XtRemoveTimeOut(f->timer);

    free_searchdata(&f);
}
	

/*
 * Check to see if the pending LDAP search operation is finished.  If
 * not, install another XtAppTimeOut and exit.  If the search is
 * complete, then either bring up an entry window or a scrolling list
 * of matches (if there was more than one match).
 */
void
check_search(f)
struct searchData	*f;
{
    int			rc, matches;
    LDAPMessage		*result = NULL, *e;
    char		errtxt[1024], info[1024];
    LDAPFiltInfo	*fi;
    char		**vals;
    char		*dn, **xdn;

    /* See if the search operation has completed */
    rc =  ldap_result(ldp, f->msgid, 1, zerotime, &result);
    switch (rc) {
    case 0:
	/* search not yet complete */
	if ((time(NULL) - f->start_time) >
	    /* Search operation timed out */
	    appConfig.ldapSearchTimeout) {
	    errorDlgEnum(f->parent, E_SEARCH_TIMED_OUT);
	    /* If there's a working dialog on-screen, destroy it */
	    destroyWorking(&(f->workingDlg));
	    if (result) {
		ldap_msgfree(result);
	    }
	    free_searchdata(&f);
	    return;
	}

	f->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
	    check_search, f);
	    return;
	    break;
    case -1:
	/* fatal error */
	f->ldaprc = ldp->ld_errno;
	if (result) {
	    ldap_msgfree(result);
	}
	if ((reconnect_ldap(f->ldaprc)) == 0) {
	    destroyWorking(&(f->workingDlg));
	    /* reconnect successful, retry */
	    do_search(f->parent, f->search_string, f->sattrs,
		  f->dattrs, f->filterspec, f->filterprefix, f->objtype,
		  f->disp_fn, f->selectWidgetName, f->options, f->cdata);
	    free_searchdata(&f);
	    return;
	}
	break;
    default:
	/* search is finished */
	info[0] = '\0';
	if (f->status) {
	    updateText(f->status, "Search completed.");
	}
	f->ldaprc = ldap_result2error(ldp, result, 0);
	if ((f->ldaprc == LDAP_SUCCESS) 
	    || (f->ldaprc == LDAP_SIZELIMIT_EXCEEDED)
	    || (f->ldaprc == LDAP_SIZELIMIT_EXCEEDED)) {
	    /* search ok, display results */

	    /* If there's a working dialog on-screen, destroy it */
	    destroyWorking(&(f->workingDlg));
	    matches = ldap_count_entries(ldp, result);
	    if (matches == 1) {
		/* Display the one match */
		e = ldap_first_entry(ldp, result);
		vals = ldap_get_values(ldp, e, "objectClass");
		dn = ldap_get_dn(ldp, e);
		f->disp_fn(f->parent, ldp, dn, vals, f->status, f->cdata);
		free(dn);
		ldap_msgfree(result);
		ldap_value_free(vals);
		free_searchdata(&f);
		return;
	    } else if (matches > 1) {
		/* Multiple matches, display scrolling list */
		if (f->ldaprc == LDAP_SIZELIMIT_EXCEEDED) {
		    strcpy(info, appConfig.emSizeLimit);
		} else if (f->ldaprc == LDAP_TIMELIMIT_EXCEEDED) {
		    strcpy(info, appConfig.emTimeLimit);
		}
		do_selection_list(f->parent, result, matches, f->dattrs,
		    f->search_string, info, (f->currfilter == NULL) ?
		    findData->findAttrP->sa_attrlabel : f->currfilter->lfi_desc,
		    f->disp_fn, f->selectWidgetName, f->options, f->cdata);
		ldap_msgfree(result);
		free_searchdata(&f);
		return;
	    } else if (matches == 0) {
		/* 
		 * No matches, go on to next filter, if any.
		 * f->ldfp will be NULL if we're dealing with a literal
		 * filter passed in, so this should just work.
		 */
		if ((!f->ldfp) || (fi = ldap_getnextfilter(f->ldfp)) == NULL) {
		    /* No more filters, report failure to user */
		    xdn = ldap_explode_dn(f->searchbase, 1);
		    sprintf(errtxt, "No \"%s\" matches for \"%s\" in %s, %s",
			f->objtype, f->search_string,
			xdn[0] ? xdn[0] : appConfig.theWorldString,
			xdn[1] ? xdn[1] : "");
		    errorDlg(f->parent, errtxt);
		    ldap_msgfree(result);
		    free_searchdata(&f);
		    ldap_value_free(xdn);
		    return;
		} else {
		    f->currfilter = fi;
		    /* Go on to the next filter */
		    f->msgid = ldap_search(ldp, f->searchbase, f->scope,
			fi->lfi_filter, f->allattrs, 0);
		    f->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
			check_search, f);
		    ldap_msgfree(result);
		    return;
		}
	    }
	}
	break;
    } /* end case */

    /* If there's a working dialog on-screen, destroy it */
    destroyWorking(&(f->workingDlg));
    /* some kind of an error occurred */
    sprintf(errtxt, "Error while searching: %s",
	ldap_err2string(f->ldaprc));
    errorDlg(f->parent, errtxt);
    return;
}



/*
 * Callback function for find dialog's 'ok' button.  Initiates a search
 * at the current searchbase.
 *
 * 10/12/95 - we're going to overload the "filterspec" argument as
 * follows:
 * - if filterspec is NULL, then then the setting of the find dialog is used.
 * - if the SRCH_OPT_LITERALFILT bit is set, then "filterspec" is taken
 *   as a literal search filter.  The filterprefix is ignored.
 * - if SRCH_OPT_LITERALFILT is *not* set, then "filterspec" is interpreted
 *   as the name of an LDAP filter tag.
 */
void
do_search(w, buf, attrs, dattrs, filterspec, filterprefix, objtype, disp_fn,
	  selectWidgetName, options, cdata) 
Widget		w;
char		*buf;		/* what we're searching for */
char		**attrs;	/* attrs to retrieve */
char		**dattrs;	/* attrs to retrieve for disambiguation */
char		*filterspec;	/* filter tag to use */
char		*filterprefix;	/* filter prefix to use */
char		*objtype;	/* descriptive obj type name */
void		(*disp_fn)();	/* function to call if single match (or
				   multiple matches and user selects one */
char		*selectWidgetName;	/* X rsrc name of selection widget */
unsigned long	options;	/* bitmap of options */
void		*cdata;		/* data to pass back to client at end */
{
    searchData		*f;
    char		errstr[1024];
    LDAPFiltInfo	*fi;
    static char		sfilter[LDAP_FILT_MAXSIZ];
    char		*filter;
    int			i, j;

    if ((buf == NULL) || (strlen(buf) == 0)) {
	errorDlg(w, "Please specify something to search for");
	return;
    }

    /* set up our per-search struct */
    f = get_searchdata();
    f->searchbase = strdup(search_base);
    f->workingDlg = (Widget) NULL;
    f->scope = findData->findObjectP->so_defaultscope;
    f->parent = w;
    f->status = findData->status;
    f->search_string = strdup(buf);
    f->etype = NULL;
    f->ldfp = NULL;
    f->start_time = time(NULL);
    f->disp_fn = disp_fn;
    f->sattrs = ccopy(attrs);
    f->dattrs = ccopy(dattrs);
    f->selectWidgetName = selectWidgetName;
    f->options = options;
    f->filterprefix = filterprefix;
    f->objtype = objtype;
    f->cdata = cdata;
    if (filterspec == NULL) {  /* Use setup of find dialog */
	f->currfilter = NULL;
	ldap_build_filter(sfilter, LDAP_FILT_MAXSIZ,
		findData->findMatchP->sm_filter,
		findData->findObjectP->so_filterprefix, NULL,
		findData->findAttrP->sa_attr, buf, NULL);
	filter = sfilter;
    } else {	/* "fewer choices" case */
	f->filterspec = strdup(filterspec);
	if (options & SRCH_OPT_LITERALFILT) {
	    ldap_build_filter(sfilter, LDAP_FILT_MAXSIZ,
		    f->filterspec, NULL, NULL, NULL, 
		    buf, NULL);
	    filter = sfilter;
	} else {
	    /* filterspec give the name of an LDAP filter tag */
	    if ((f->ldfp = ldap_init_getfilter(appConfig.ldapFilterFile))
			== NULL) {
		sprintf(errstr, "Unable to open LDAP filter file %s",
		    appConfig.ldapFilterFile);
		errorDlg(w, errstr);
		/* XXX clean up */
		return;
	    }
	    if (options & SRCH_OPT_PEOPLEONLY) {
		ldap_setfilteraffixes(f->ldfp, "(&objectClass=Person)", ")");
	    } else {
		if (filterprefix) {
		    ldap_setfilteraffixes(f->ldfp, filterprefix, ")");
		}
	    }
	    fi = ldap_getfirstfilter(f->ldfp, filterspec, f->search_string);
	    if (fi == NULL) {
		sprintf(errstr, "No search filter %s found in file %s",
		    filterspec, appConfig.ldapFilterFile);
		errorDlg(w, errstr);
		ldap_setfilteraffixes(f->ldfp, NULL, NULL);
		/* XXX Clean up */
		return;
	    }
	    f->currfilter = fi;
	    filter = fi->lfi_filter;
	    f->scope = fi->lfi_scope;
	}
    }

    if (f->status) {
	updateText(f->status, "Searching...");
    }
	
    /* Assemble all the attributes into one list */
    for (i = 0; f->sattrs && f->sattrs[i] != NULL; i++);
    for (j = 0; f->dattrs && f->dattrs[j] != NULL; j++, i++);
    i++;
    f->allattrs = (char **) malloc(i * sizeof(char *));
    memset(f->allattrs, 0, i * sizeof(char *));
    for (i = 0; f->sattrs && f->sattrs[i] != NULL; i++) {
	f->allattrs[i] = strdup(f->sattrs[i]);
    }
    for (j = 0; f->dattrs && dattrs[j] != NULL; j++, i++) {
	f->allattrs[i] = strdup(f->dattrs[j]);
    }
    /* fire off the search */
    f->msgid = ldap_search(ldp, f->searchbase, f->scope,
	filter, f->allattrs, 0);
    if (f->msgid == -1) {
	/* Error */
	if ((reconnect_ldap(f->ldaprc)) == 0) {
	    /* reconnect successful, retry */
	    do_search(f->parent, f->search_string, f->sattrs, f->dattrs,
		f->filterspec, f->filterprefix, f->objtype, f->disp_fn,
		f->selectWidgetName, f->options, f->cdata);
	} else {
	    sprintf(errstr, "Error while searching: %s",
		ldap_err2string(ldp->ld_errno));
	    errorDlg(f->parent, errstr);
	    /* XXX clean up */
	}
	return;
    }

    /* Set up an Xt App timeout function to handle the result */
    f->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL, check_search, f);
	
    /* Display a working dialog */
    f->workingDlg = create_working_dialog(w, "Searching...", "Cancel",
	NULL, NULL, cancel_search, NULL, NULL, f, w);
    /* Disable the "Done" button */
    /* XXX */
}







/*
 *
 * UFN search
 */
void
do_ufn_search()
{
}




void
verify_dn()
{
}
