/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Display entries found in do_read()
 */

#include <Xm/SelectioB.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/MessageB.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/ScrolledW.h>

#include <lber.h>
#include <ldap.h>

#include <stdio.h>
#include <sys/time.h>

#include "dtypes.h"
#include "util.h"
#include "menu.h"
#include "globals.h"
#include "select.h"
#include "read.h"
#include "bitmaps.h"
#include "dialog.h"
#include "help.h"

#define	SELECTION_SORT_BY_DN		1
#define	SELECTION_SORT_BY_SURNAME	2


/* forward references */
#ifndef NEEDPROTOS
void selection_made_cb();
void selection_ok_cb();
void selection_list_destroy();
void createItemListDialog();
void sortSelectionItems();
void selectionToggleSortMode();
#else /* !NEEDPROTOS */
void selection_made_cb(Widget w, XtPointer client_data,
	    XmAnyCallbackStruct *cbs);
void selection_ok_cb(Widget w, XtPointer client_data,
	    XmListCallbackStruct *cbs);
void selection_list_destroy(Widget w, XtPointer client_data,
	    XmAnyCallbackStruct *cbs);
void createItemListDialog(Widget w, selectList *selectionItems, int count,
	    selectData *s, char *widgetname, char *name, char *lfi_desc,
	    char *info);
void sortSelectionItems(selectList *selectionItems, int count, int sortType);
void selectionToggleSortMode(Widget w, selectData *s, XmAnyCallbackStruct *cbs);
#endif /* !NEEDPROTOS */

/*
 * Present a scrolling list of entries, and allow the user to select
 * one (maybe allow multiple selections in the future).
 *
 * Note: caller will free result, attrs, name and lfi_desc after this
 *       routine returns.  Copy them if you need them.
 *
 * The last argument, widgetname, allows specification at runtime of the
 * widget's resource name.  This allows us to use different labels for
 * the buttons, etc. depending on if the selection dialog was brought up
 * as a result of a find or an authentication operation.
 */

void
do_selection_list(w, result, count, attrs, name, info, lfi_desc, disp_fn,
		  widgetname, options, cdata)
Widget		w;
LDAPMessage	*result;	/* result of multiple-match ldap search */
int		count;		/* number of matches */
char		**attrs;	/* disambiguation attributes */
char		*name;		/* name we searched for */
char		*info;		/* extra info to display in title bar */
char		*lfi_desc;	/* filter description */
void		(*disp_fn)();	/* function to call when user selects */
char		*widgetname;	/* name of X resource for widget */
unsigned long	options;
void		*cdata;		/* data to pass back to disp_fn() */
{
#define	MAX_LIST_ELT_SIZE	16384 /* XXX */

    selectList		*selectionItems;
    char		**surnames;
    selectData		*s;
    LDAPMessage		*e;
    int			i, j;
    char		*dn, **xdn;
    XmString		xmstmp, xmssep, xmstmp2, xmsdata;
    char		buf[MAX_LIST_ELT_SIZE];
    char		**p;
    int			got_surnames = 0;

    /* Allocate the main data structure for our list */
    selectionItems = (selectList *) malloc((count) * sizeof(selectList));
    memset(selectionItems, 0, (count) * sizeof(selectList));

    /* Allocate the selectData structure */
    s = (selectData *) malloc(sizeof(selectData));
    s->selectionItems = selectionItems;
    s->disp_fn = disp_fn;
    s->options = options;
    s->cdata = cdata;
    s->count = count;
    s->dn_list = NULL;
    s->objcllist = NULL;
    s->sort_mode = SELECTION_SORT_BY_DN;
    if (!lfi_desc) {
	lfi_desc = "";
    }

    /* Retrieve the results */
    for (e = ldap_first_entry(ldp, result), i = 0; e != NULL;
	    e = ldap_next_entry(ldp, e), i++) {

	/* Save the DN */
	selectionItems[i].dn = ldap_get_dn(ldp, e);

	/* Save the object class list */
	selectionItems[i].objclasses = ldap_get_values(ldp, e, "objectClass");

	/* Save the surname, if any */
	surnames = ldap_get_values(ldp, e, "sn");
	if (surnames != NULL) {
	    got_surnames = 1;
	    selectionItems[i].surname = strdup(*surnames);
	    ldap_value_free(surnames);
	}

	/* Put the leftmost component of the dn in the scrolling list */
	dn = ldap_get_dn(ldp, e);
	xdn = ldap_explode_dn(dn, 1);
	sprintf(buf, "%s\n", xdn[0]);
	xmstmp = XmStringSegmentCreate(buf, "selectfont",
		XmSTRING_DIRECTION_L_TO_R,
		False);
	free(dn);
	ldap_value_free(xdn);
	xmssep = XmStringSeparatorCreate();

	/* Get the values for the list */
	for (j = 0; attrs[j] != NULL; j++) {
	    if (!strcmp(attrs[j], "objectClass")) {
		continue;
	    }
	    xmstmp2 = XmStringCopy(xmstmp);
	    XmStringFree(xmstmp);
	    xmstmp = XmStringConcat(xmstmp2, xmssep);
	    XmStringFree(xmstmp2);
	    p = ldap_get_values(ldp, e, attrs[j]);
	    if (p != NULL) {
		/*
		 * If really long, truncate string and add an ellipsis.
		 */
		if (strlen(*p) > appConfig.maxSelLen) {
		    char *el = &(*p)[appConfig.maxSelLen - 4];
		    *el++ = '.';
		    *el++ = '.';
		    *el++ = '.';
		    *el = '\0';
		}
		sprintf(buf, "%s", *p);
		xmsdata =  XmStringSegmentCreate(buf, "selectfont2",
			XmSTRING_DIRECTION_L_TO_R, False);
		ldap_value_free(p);
		xmstmp2 = XmStringCopy(xmstmp);
		XmStringFree(xmstmp);
		xmstmp = XmStringConcat(xmstmp2, xmsdata);
		XmStringFree(xmstmp2);
		XmStringFree(xmsdata);
	    }
	}
	selectionItems[i].xmstr = xmstmp;
	XmStringFree(xmssep);
    }

    sortSelectionItems(selectionItems, count, SELECTION_SORT_BY_DN);

    /* Use RDN if this is a DN (since the name would be too long */
    if (is_dn(name) && strlen(name) > 20) {
	/* Use rdn instead of DN, because it's too long */
	char **xdn;
	xdn = ldap_explode_dn(name, 1);
	strcpy(buf, xdn[0]);
	ldap_value_free(xdn);
    } else {
	strcpy(buf, name);
    }

    createItemListDialog(w, selectionItems, count, s, widgetname, buf,
	    lfi_desc, info);
    if (!got_surnames) {
	XtSetSensitive(s->sortButton, False);
    }
}




void
createItemListDialog(w, selectionItems, count, s, widgetname, name, lfi_desc,
	info)
Widget			w;
selectList		*selectionItems;
int			count;
selectData		*s;
char			*widgetname;
char			*name;
char			*lfi_desc;
char			*info;
{
    XmStringTable	list_items;
    int			i, n;
    char		buf[MAX_LIST_ELT_SIZE];
    Widget		selectionDlg, selectionForm, selectionInfo,
			selectionList, selectionSortButton;
    XmString		str;
    Arg			args[15];

    /* Allocate the XmStrings for our list */
    list_items = (XmStringTable) XtMalloc((count) * sizeof(XmString *));

    /* Copy sorted list items */
    for (i = 0; i < count; i++) {
	list_items[i] = selectionItems[i].xmstr;
    }

    /* Create the dialog */
    n = 0;
    sprintf(buf, "Matches for \"%s\"", name);   /* XXX resource */
    XtSetArg(args[n], XmNtitle, buf); n++;
    selectionDlg = XmCreateMessageDialog(w,
	    (widgetname) ? widgetname : "selectDlg", args, n);
    s->selectionDlg = selectionDlg;
    s->okButton = XmMessageBoxGetChild(selectionDlg, XmDIALOG_OK_BUTTON);

    selectionForm = XtVaCreateManagedWidget("selectionForm",
	    xmFormWidgetClass,
	    selectionDlg,
	    NULL);
    selectionSortButton = XtVaCreateManagedWidget("selectionSortButton",
	    xmPushButtonWidgetClass,
	    selectionForm,
	    XmNtopAttachment, XmATTACH_FORM,
	    XmNleftAttachment, XmATTACH_FORM,
	    NULL);
    XtAddCallback(selectionSortButton, XmNactivateCallback,
	    (XtCallbackProc) selectionToggleSortMode, s);
    s->sortButton = selectionSortButton;
    /* Status area */
    str = XmStringCreateLtoR(" ", XmSTRING_DEFAULT_CHARSET);
    s->status = XtVaCreateManagedWidget("FindStatus", xmLabelWidgetClass,
	    selectionForm,
	    XmNlabelString, str,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, selectionSortButton,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_FORM,
	    NULL);
    XmStringFree(str);

    n = 0;
    sprintf(buf, "%d %s matches for \"%s\"", count, lfi_desc, name);
    if (info) {
	strcat(buf, "\n");
	strcat(buf, info);
    }
    str = XmStringCreateLtoR(buf, XmSTRING_DEFAULT_CHARSET);
    selectionInfo = XtVaCreateManagedWidget("selectionInfo",
	    xmLabelGadgetClass, selectionForm,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, s->status,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNlabelString, str,
	    NULL);
    XmStringFree(str);
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, selectionInfo); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNscrollBarDisplayPolicy, XmSTATIC); n++;
    XtSetArg(args[n], XmNlistSpacing, 5); n++;
    XtSetArg(args[n], XmNvisibleItemCount, (count >= 10) ? 10 : count); n++;
    XtSetArg(args[n], XmNitemCount, count); n++;
    XtSetArg(args[n], XmNitems, list_items); n++;
    XtSetArg(args[n], XmNselectionPolicy, XmSINGLE_SELECT); n++;
    selectionList = XmCreateScrolledList(selectionForm, "selectionList",
					     args, n);
    /*
     * XXX we should really support XmMULTIPLE_SELECT, but there
     * is a bug which causes a core dump is multiple items are
     * selected and displayed.
     */
    s->selectionList = selectionList;
    XmListSelectPos(selectionList, 1, False);
    /* SetFocus(selectionDlg, selectionList); */
		
    /*
     * Point at our selectData structure
     */
    XtVaSetValues(selectionList, XmNuserData, s, NULL);
    XtVaSetValues(selectionDlg, XmNuserData, s, NULL);
    XtVaSetValues(s->okButton, XmNuserData, s, NULL);

    XtManageChild(selectionList);
    XtAddCallback(selectionList, XmNdefaultActionCallback,
	    (XtCallbackProc) selection_made_cb, s);
    XtAddCallback(selectionDlg, XmNcancelCallback,
	    (XtCallbackProc) selection_list_destroy, NULL);
    XtAddCallback(selectionDlg, XmNokCallback,
	    (XtCallbackProc) selection_ok_cb, s);
    XtAddCallback(selectionDlg, XmNhelpCallback,
	    (XtCallbackProc) doHelpDlg, HLP_TOPIC_SELECT_LIST);
    XtManageChild(selectionDlg);
    XtPopup(XtParent(selectionDlg), XtGrabNone);

    /* Free the XmString pointers - they were copied by the list widget */
    XtFree((XtPointer) list_items);

    if (s->sort_mode == SELECTION_SORT_BY_SURNAME) {
	str = XmStringCreateLocalized(appConfig.msgSortByFullName);
    } else {
	str = XmStringCreateLocalized(appConfig.msgSortBySurname);
    }
    XtVaSetValues(s->sortButton, XmNlabelString, str, NULL);
    XmStringFree(str);

}



static int
compare_sel_dns(a, b)
selectList	*a;
selectList	*b;
{
    return(strcasecmp(a->dn, b->dn));
}

static int
compare_sel_sns(a, b)
selectList	*a;
selectList	*b;
{
    return(strcasecmp(a->surname, b->surname));
}

/*
 * Sort an array of selectionItems.
 */
void
sortSelectionItems(selectionItems, count, sortType)
selectList	*selectionItems;
int		count;
int		sortType;
{
    if (sortType == SELECTION_SORT_BY_SURNAME) {
	qsort((char *) selectionItems, count, sizeof(selectList),
		compare_sel_sns);
    } else {
	qsort((char *) selectionItems, count, sizeof(selectList),
		compare_sel_dns);
    }
}
 
void
selectionToggleSortMode(w, s, cbs)
Widget			w;
selectData		*s;
XmAnyCallbackStruct	*cbs;
{
    XmStringTable	list_items;
    XmString		str;
    int			i;

    if (s->sort_mode == SELECTION_SORT_BY_SURNAME) {
	sortSelectionItems(s->selectionItems, s->count,
		SELECTION_SORT_BY_DN);
	s->sort_mode = SELECTION_SORT_BY_DN;
    } else {
	sortSelectionItems(s->selectionItems, s->count,
		SELECTION_SORT_BY_SURNAME);
	s->sort_mode = SELECTION_SORT_BY_SURNAME;
    }

    /* Allocate the XmStrings for our list */
    list_items = (XmStringTable) XtMalloc((s->count) * sizeof(XmString *));

    /* Copy sorted list items */
    for (i = 0; i < s->count; i++) {
	list_items[i] = s->selectionItems[i].xmstr;
    }

    XtVaSetValues(s->selectionList,
	    XmNitemCount, s->count,
	    XmNitems, list_items,
	    NULL);

    XtFree((XtPointer) list_items);

    if (s->sort_mode == SELECTION_SORT_BY_SURNAME) {
	str = XmStringCreateLocalized(appConfig.msgSortByFullName);
    } else {
	str = XmStringCreateLocalized(appConfig.msgSortBySurname);
    }
    XtVaSetValues(s->sortButton, XmNlabelString, str, NULL);
    XmStringFree(str);
}
	



/* 
 * Called when user double-clicks an item in the scrolling list.
 * For each selected item, retrieve the entry and display
 * it in a window.
 */
void
selection_made_cb(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    selectData	*s = (selectData *) client_data;

    if (cbs->event->type == KeyPress) {
	/*
	 * We need to ignore keypress events, since they will also
	 * cause selection_ok_cb to be called automatically.  If
	 * we don't ignore it, selection_ok_cb gets called twice.
	 */
	return;
    }

    /* This is a little sleazy, but, what the hell... */
    XtCallCallbacks(s->selectionDlg, XmNokCallback, NULL);
}





/* 
 * Called when user clicks on the Ok button
 * For each selected item, retrieve the entry and display
 * it in a window.
 */
void
selection_ok_cb(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmListCallbackStruct	*cbs;
{
    int			i;
    XmString		t;
    selectData		*s;
    int			selected_item_count, *selected_item_positions;
    Widget		status;
    char		*dn;
    char		**cl;
    Widget		rparent;
    Boolean		displayed = False;

    XtVaGetValues(w, XmNuserData, &s, NULL);

    if (XmListGetSelectedPos(s->selectionList, &selected_item_positions,
	    &selected_item_count)) {

	if (s->status) {
	    t = XmStringCreateLocalized("Reading...");
	    XtVaSetValues(s->status, XmNlabelString, t, NULL);
	    XmStringFree(t);
	}
	for (i = 0; i < selected_item_count; i++) {
	    dn = s->selectionItems[selected_item_positions[i] - 1].dn;
	    cl = s->selectionItems[selected_item_positions[i] - 1].objclasses;
	    if (!(s->options & SRCH_OPT_PERSIST)) {
		/* Dialog will diappear soon */
		rparent = toplevel;
		status = NULL;
	    } else {
		rparent = w;
		status = s->status;
	    }
	    displayed = True;
	    s->disp_fn(rparent, ldp, dn, cl,  status, s->cdata);
	}
	XtFree((char *) selected_item_positions);
    } else {
	/* Nothing selected */
    }
    if (displayed && !(s->options & SRCH_OPT_PERSIST)) {
	XtCallCallbacks(s->selectionDlg, XmNcancelCallback, NULL);
    }
}


/*
 * Clean up.
 */
void
selection_list_destroy(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    selectData		*s;
    int			i;

    XtVaGetValues(w, XmNuserData, &s, NULL);
	
    if (s->selectionItems != NULL) {
	for (i = 0; i < s->count; i++) {
	    if (s->selectionItems[i].dn != NULL) {
		free(s->selectionItems[i].dn);
	    }
	    if (s->selectionItems[i].surname != NULL) {
		free(s->selectionItems[i].surname);
	    }
	    if (s->selectionItems[i].objclasses != NULL) {
		ldap_value_free(s->selectionItems[i].objclasses);
	    }
	    if (s->selectionItems[i].xmstr != NULL) {
		XmStringFree(s->selectionItems[i].xmstr);
	    }
	}
	free(s->selectionItems);
    }
    if (s->dn_list) {
	for (i = 0; s->dn_list[i] != NULL; i++) {
	    free(s->dn_list[i]);
	}
	free(s->dn_list);
    }
    if (s->objcllist) {
	for (i = 0; s->objcllist[i] != NULL; i++) {
	    ldap_value_free(s->objcllist[i]);
	}
	free(s->objcllist);
    }
    free(s);
    XtDestroyWidget(w);
}
