/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Notes on converting strings to bervals
 * --------------------------------------
 *
 * When writing out text attributes:
 *
 * CIS       - any newlines signal the start of a new value.  Extraneous
 *             newlines are tossed.
 *
 * MLS       - two or more consecutive newlines signal the start of a new
 *             value.  Lines separated by exactly one newline are joined
 *             with " $ " and written as a single value.
 *
 * MLS, 1val - always written as a single value.  Each newline is replaced
 *             with " $ ".  Blank lines are replaced with "\20 $ "
 *
 * Example: if the user has typed the following:
 *
 * This is line 1.
 * This is line 2.
 *
 * This is line 3.
 *
 *
 * This is line 4.
 * <end of example>
 *
 * Then the following bervals would be generated if the string is interpreted
 * as:
 *
 * 1) a CIS:
 * bv[0]: "This is line 1."
 * bv[1]: "This is line 2."
 * bv[2]: "This is line 3."
 * bv[3]: "This is line 4."
 *
 * 2) an MLS (but not 1val)
 * bv[0]: "This is line 1. $ This is line 2."
 * bv[1]: "This is line 3."
 * bv[2]: "This is line 4."
 *
 * 3) an MLS (1val)
 * bv[0]: "This is line 1. $ This is line 2. $ \20 $ This is line 3. $ \20 $ \20 $ This is line 4."
 */

/*
 * Syntax handlers for various data types.
 */

#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <fcntl.h>
#include <time.h>
#include <malloc.h>
#include <string.h>
#include <signal.h>

#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleBG.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/ScrolledW.h>
#include <Xm/PanedW.h>
#include <Xm/SeparatoG.h>
#include <Xm/ArrowB.h>
#include <Xm/Frame.h>
#include <Xm/Protocols.h>
#include <Xm/DragDrop.h>

#include <lber.h>
#include <ldap.h>

#include "disptmpl.h"
#include "dtypes.h"
#include "util.h"
#include "menu.h"
#include "globals.h"
#include "display.h"
#include "syn.h"
#include "bitmaps.h"
#include "dialog.h"
#include "browse.h"
#include "write.h"
#include "edit.h"
#include "find.h"
#include "errtxt.h"

/* Forward references */
#ifndef NEEDPROTOS
static long		gtime();
#else /* !NEEDPROTOS */
static long		gtime(struct tm *tm);
#endif /* !NEEDPROTOS */

/*
 * Convert a berval to a char *.  Converts " $ " to \n iff converting
 * a multi-line string.  The returned string needs to be freed by
 * the caller.
 */
void *
convToText(vals, flags)
struct berval		**vals;
unsigned int		flags;		/* non-zero if this is a multi-line string */
{
#define	IS_MLS	1
#define	ONEVAL	2
    int		i, len, di, si, maxval;
    char	*d, *s;

    if (vals !=  NULL) {
	len = maxval = 0;
	for (i = 0; vals[i] != NULL; i++) {
	    len += vals[i]->bv_len + 1;	/* allow space for \n */
	    maxval = i;
	}
	len++;	/* ending NULL */
	maxval--;
	if ((d = (char *) malloc(len)) == NULL) {
	    fprintf(stderr, "out of memory, convToText\n");
	    exit(1);
	}
	memset(d, 0, len);
	di = 0;
	for (i = 0; vals[i] != NULL; i++) {
	    s = vals[i]->bv_val;
	    si = 0;
	    while (si < vals[i]->bv_len) {
		if (s[si] == '$') {
		    if (s[si + 1] == ' ') {
			si++;
		    }
		    /* Back up over spaces from ' $ ' */
		    if (di > 1) {
			di--;
		    }
		    /*
		    while (di > 1 && isspace(d[di - 1])) {
			di--;
		    }
		    */
		    if (flags & IS_MLS) {
			d[di] = '\n';
		    }
		} else {
		    d[di] = s[si];
		}
		di++;
		si++;
	    }
	    if (i <= maxval) {
		d[di++] = '\n';
	    }
	}
	d[di] = '\0';
	return((void *) d);
    }
}


/*
 * Convert a multi-line string to a char *.
 */
void *
syn_MLSToText(vals)
struct berval	**vals;
{
    return(convToText(vals, IS_MLS));
}


/*
 * Convert a Case Ignore String to a char *.
 */
void *
syn_CISToText(vals)
struct berval	**vals;
{
    return(convToText(vals, 0));
}



#define GET2BYTENUM( p )	(( *p - '0' ) * 10 + ( *(p+1) - '0' ))

static char *
time2text(ldtimestr, dateonly)
char *ldtimestr;
int dateonly;
{
    struct tm		t;
    char		*p, *timestr, zone, *fmterr = "badly formatted time";
    time_t		gmttime;

    memset( (char *)&t, 0, sizeof( struct tm ));
    if ( strlen( ldtimestr ) < 13 ) {
	return( fmterr );
    }

    for ( p = ldtimestr; p - ldtimestr < 12; ++p ) {
	if ( !isdigit( *p )) {
	    return( fmterr );
	}
    }

    p = ldtimestr;
    t.tm_year = GET2BYTENUM( p ); p += 2;
    t.tm_mon = GET2BYTENUM( p ) - 1; p += 2;
    t.tm_mday = GET2BYTENUM( p ); p += 2;
    t.tm_hour = GET2BYTENUM( p ); p += 2;
    t.tm_min = GET2BYTENUM( p ); p += 2;
    t.tm_sec = GET2BYTENUM( p ); p += 2;

    if (( zone = *p ) == 'Z' ) {	/* GMT */
	zone = '\0';	/* no need to indicate on screen, so we make it null */
    }

    gmttime = gtime( &t );
    timestr = ctime( &gmttime );

    timestr[ strlen( timestr ) - 1 ] = zone;	/* replace trailing newline */
    if ( dateonly ) {
	strcpy( timestr + 11, timestr + 20 );
    }

    return( strdup( timestr ) );
}


/* gtime(): the inverse of localtime().
	This routine was supplied by Mike Accetta at CMU many years ago.
 */

int	dmsize[] = {
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

#define	dysize(y)	\
	(((y) % 4) ? 365 : (((y) % 100) ? 366 : (((y) % 400) ? 365 : 366)))

#define	YEAR(y)		((y) >= 100 ? (y) : (y) + 1900)

static long
gtime(tm)
struct tm *tm;
{
    register int    i,
                    sec,
                    mins,
                    hour,
                    mday,
                    mon,
                    year;
    register long   result;

    if ((sec = tm -> tm_sec) < 0 || sec > 59
	    || (mins = tm -> tm_min) < 0 || mins > 59
	    || (hour = tm -> tm_hour) < 0 || hour > 24
	    || (mday = tm -> tm_mday) < 1 || mday > 31
	    || (mon = tm -> tm_mon + 1) < 1 || mon > 12)
	return ((long) -1);
    if (hour == 24) {
	hour = 0;
	mday++;
    }
    year = YEAR (tm -> tm_year);

    result = 0L;
    for (i = 1970; i < year; i++)
	result += dysize (i);
    if (dysize (year) == 366 && mon >= 3)
	result++;
    while (--mon)
	result += dmsize[mon - 1];
    result += mday - 1;
    result = 24 * result + hour;
    result = 60 * result + mins;
    result = 60 * result + sec;

    return result;
}



struct berval **
syn_FromTime()
{
    return(NULL);
}


struct berval **
syn_FromDate()
{
    return(NULL);
}



/*
 * Convert a time or date berval to a char *.  Returned string must be
 * freed by the caller.
 */
void *
convToTime(vals, is_date)
struct berval	**vals;
Boolean		is_date;
{
    int	i, nv, len;
    char	**cvals = NULL;
    char	*ret;

    if (vals !=  NULL) {
	len = 0;
	for (nv = 0; vals[nv] != NULL; nv++);
	cvals = (char **) malloc((nv + 1) * sizeof(char *));
	for (i = 0; i < nv; i++) {
	    cvals[i] = time2text(vals[i]->bv_val, is_date);
	    len += strlen(cvals[i]) + 1;
	}
	cvals[nv] = NULL;
	ret = (char *) malloc(len + 1);
	ret[0] = '\0';
	for (i = 0; i < nv; i++) {
	    strcat(ret, cvals[i]);
	    strcat(ret, "\n");
	    free(cvals[i]);
	}
	ret[len - 1] = '\0';
	free(cvals);
	return((void *) ret);
    }
    return((void *) strdup(""));
}



void *
syn_ToDate(vals)
struct berval **vals;
{
    return(convToTime(vals, True));
}



void *
syn_ToTime(vals)
struct berval **vals;
{
    return(convToTime(vals, False));
}



struct berval **
syn_FromURL()
{
}

void *
syn_ToURL()
{
    return NULL;
}

struct berval **
syn_FromMember()
{
}

void *
syn_ToMember()
{
}

/*
 * Convert a buffer of text to a null-terminated array of bervals.
 * Two or more newlines in a row signal the start of a new value.
 * Three or more newlines are the same as two newlines.
 * Leading or trailing sequences of newlines are ignored.
 */
struct berval **
convFromText(text, flags)
char		*text;
unsigned int	flags;
{
    char		*d, *p;
    int			nvals = 0;
    char		obuf[8192];
    struct berval	**bvs, **tmp;

    if ((text == NULL) || (strlen(text) == 0)) {
	return NULL;
    }

    p = text;
    /* skip any leading newlines */
    while (*p && *p == '\n') {
	p++;
    }
    d = obuf;
    bvs = (struct berval **) malloc(sizeof(struct berval *));
    for (;;) {
	if ((!*p) || (*p == '\n')) {
	    /*
	     * only make a new berval if:
	     * 1) we've run out of chars, or
	     * 2) not processing a multiline string, or
	     * 3) consecutive newlines found and not ONEVAL
	     */

	    if ((!*p) || (!(flags & IS_MLS)) ||
		    ((*(p + 1) == '\n') && (!(flags & ONEVAL)))) {
tieup:		*d = '\0';
		tmp = (struct berval **) realloc(bvs,
			  (nvals + 2) * sizeof(struct berval *));
		bvs = tmp;
		bvs[nvals] = (struct berval *) malloc(sizeof(struct berval));
		bvs[nvals]->bv_len = d - obuf;
		bvs[nvals]->bv_val = strdup(obuf);
		nvals++;
		while (*p && *++p == '\n');	/* skip */
		if (!*p) {
		    break;
		}
		d = obuf;
	    } else {
		/* MLS case, 1val */
		*d++ = ' ';
		*d++ = '$';
		*d++ = ' ';
		p++;
		for (;;) {
		    if (!*p) {
			goto tieup;
		    } else if (*p == '\n') {
			p++;
			*d++ = '\\';
			*d++ = '2';
			*d++ = '0';
			*d++ = ' ';
			*d++ = '$';
			*d++ = ' ';
		    } else {
			break;
		    }
		}
	    }
	} else {
	    *d++ = *p++;
	}
    }
    bvs[nvals] = NULL;
    return(bvs);
}



struct berval **
syn_TextToMLS(val)
char	*val;
{
    return(convFromText(val, IS_MLS));
}


/*
 * Convert a char * to a multi-line string, preserving order.  Used
 * when converting an MLS that is marked "1val".  Used when the ordering
 * of text is important, e.g. when there are several paragraphs of text in
 * an MLS and the order should be preserved.
 */
struct berval **
syn_TextToMLS1val(val)
char	*val;
{
    return(convFromText(val, IS_MLS | ONEVAL));
}



struct berval **
syn_TextToCIS(val)
char	*val;
{
    return(convFromText(val, 0));
}


void *
syn_RFC822ToText(vals)
struct berval **vals;
{
    return(convToText(vals, 0));
}

struct berval **
syn_TextToRFC822(val)
char *val;
{
    return(convFromText(val, 0));
}

int
syn_VerifyRFC822(val)
char *val;
{
    if (strchr(val, ' ') != NULL) {
	return(E_RFC822_BLANKS_DISALLOWED);
    } else {
	return(0);
    }
}




/*
 * Convert a a series of newline-separated DNs to a bervec.
 */
struct berval **
syn_FromDN(val)
void *val;
{
    int		nvals;
    int		i;
    char		*p, *s;
    struct berval	**bvs;

    if (val == NULL) {
	return(NULL);
    }

    p = val;
    nvals = 0;
    while (*p) {
	if (*p == '\n') {
	    nvals++;
	}
	p++;
    }
    /* did we end with \n? */
    if ((p == val) || (*--p != '\n')) {
	nvals++;
    }
    bvs = (struct berval **) malloc((nvals + 1) * sizeof(struct berval *));

    p = val;
    i = 0;
    while (*p) {
	/* find end of current line */
	for (s = p; *p && (*p != '\n'); p++);
	if (s == p) {
	    if (*p) {
		p++;
	    }
	    continue;
	}
	bvs[i] = (struct berval *) malloc(sizeof(struct berval));
	bvs[i]->bv_len = p - s;
	if (*p == '\n') {
	    p++;
	}
	bvs[i]->bv_val = (char *) malloc(bvs[i]->bv_len);
	memcpy(bvs[i]->bv_val, s, bvs[i]->bv_len);
	i++;
    }
    bvs[nvals] = NULL;
    return(bvs);
}




/*
 * Convert a bervec of DNs to a newline-separated char *.
 */
void *
syn_ToDN(vals)
struct berval **vals;
{
#ifdef notdef
    char		*obuf, *p, *d;
    int		i, j, size = 0;

    if (!vals) {
	return(NULL);
    }
    for (i = 0; vals[i] != NULL; i++) {
	size += (vals[i]->bv_len + 1);
    }
    obuf = (char *) malloc(size + 1);
    d = obuf;
    for (i = 0; vals[i] != NULL; i++) {
	for (j = 0, p = vals[i]->bv_val; j < vals[i]->bv_len; j++) {
	    *d++ = *p++;
	}
	*d++ = '\n';
    }
    *d = '\0';
    return(obuf);
#else
    return NULL;
#endif

}






/*
 * For DNs, we display an arrow button the user can click on, and the
 * leftmost component of the DN.  XXX May want to offer a way of
 * displaying the DN, or maybe copying the DN to the clipboard.
 */
void *
syn_DisplayDN(parent, dad)
Widget		parent;
displayAttrData	*dad;
{
    Widget		DNrc, DNForm;
    int			i;
    attrData		*ad = dad->dad_attrData;
    struct berval	**vals = dad->dad_attrData->ad_vals;

    /* Put vals in a RowColumn widget */
    DNrc = XtVaCreateWidget("DNrc", xmRowColumnWidgetClass,
			    parent,
			    XmNpacking, XmPACK_TIGHT,
			    XmNorientation, XmVERTICAL,
			    NULL);
    if (vals == NULL) {
	ad->ad_nvals = 0;
	dad->dad_valueWidgetList = 
		(WidgetList) malloc(sizeof(Widget));
    } else {
	for (i = 0; vals[i] != NULL; i++);
	ad->ad_nvals = i;
	dad->dad_valueWidgetList = 
		(WidgetList) malloc((ad->ad_nvals + 1) * sizeof(Widget));
	for (i = 0; vals[i] != NULL; i++) {
	    /* Create a form widget to hold the button and label */
	    DNForm = createDNWidget(DNrc, dad, vals[i]->bv_val);
	    dad->dad_valueWidgetList[i] = DNForm;
	}
    }

    dad->dad_osval = NULL; /* For DNs, we don't use ad_osval */
    dad->dad_valueWidgetList[ad->ad_nvals] = NULL;
    dad->dad_numValueWidgets = ad->ad_nvals;

    return(DNrc);
}
	


/*
 * Add a DN to the list of displayed values.
 */
int
syn_AddDN(dad, value)
displayAttrData	*dad;
char		*value;
{
    Widget		DNForm, DNLabel;
    char		*dn;
    attrData		*ad = dad->dad_attrData;
    int			i;

    /* Check for duplicate value */
    for (i = 0; i < ad->ad_nvals; i++) {
	DNLabel = getLabelGadgetChild(dad->dad_valueWidgetList[i]);
	if (DNLabel != NULL) {
	    XtVaGetValues(DNLabel, XmNuserData, &dn, NULL);
	    if (dn && !strcasecmp(value, dn)) {
		return(AD_DUPVALUE);
	    }
	}
    }

    berval_add(&(ad->ad_vals), value);
    DNForm = createDNWidget(dad->dad_valueWidget, dad,
	    ad->ad_vals[ad->ad_nvals]->bv_val);

    /* Update the list of value widgets */
    ad->ad_nvals++;
    if (dad->dad_valueWidgetList != NULL) {
	dad->dad_valueWidgetList =
		(WidgetList) realloc(dad->dad_valueWidgetList,
		(ad->ad_nvals + 1) * sizeof(Widget));
    } else {
	dad->dad_valueWidgetList = 
		(WidgetList) malloc((ad->ad_nvals + 1) * sizeof(Widget));
    }
    dad->dad_valueWidgetList[ad->ad_nvals - 1] = DNForm;
    dad->dad_valueWidgetList[ad->ad_nvals] = NULL;

    /* update flags, save button */
    dad->dad_attrData->ad_flags |= AD_DIRTY;
    updateDisplayDlgSaveButton(dad->dad_valueWidget, -1);

    return(AD_NOERR);
}



    


/*
 * Remove DNs from the on-screen display.  The widget isn't actually
 * destroyed, just unmanaged (faster).
 */
int
syn_DeleteDN(dad, value)
displayAttrData	*dad;
char		*value;
{
    Widget	saveWidget, DNLabel;
    char	*dn;
    attrData	*ad = dad->dad_attrData;
    int		i, j, found = 0;

    /* Find value to remove */
    for (i = 0; i < ad->ad_nvals; i++) {
	DNLabel = getLabelGadgetChild(dad->dad_valueWidgetList[i]);
	if (DNLabel != NULL) {
	    XtVaGetValues(DNLabel, XmNuserData, &dn, NULL);
	    if (!strcasecmp(value, dn)) {
		found = 1;
		break;
	    }
	}
    }
    if (!found) {
	return(AD_NOTFOUND);
    } else {
	if (berval_del(&(ad->ad_vals), value) < 0) {
	    errorDlg(GetTopShell(dad->dad_valueWidget),
		    "Cannot find value to delete");
	    return(AD_NOTFOUND);
	}
	XtUnmanageChild(dad->dad_valueWidgetList[i]);
	/* move to end of list */
	saveWidget = dad->dad_valueWidgetList[i];
	for (j = i; j < ad->ad_nvals; j++) {
	    dad->dad_valueWidgetList[j] = dad->dad_valueWidgetList[j + 1];
	}
	dad->dad_valueWidgetList[ad->ad_nvals - 1] = saveWidget;
	ad->ad_nvals--;

	/* update flags, save button */
	dad->dad_attrData->ad_flags |= AD_DIRTY;
	updateDisplayDlgSaveButton(dad->dad_valueWidget, -1);
    }
    return(AD_NOERR);
}






/*
 * Display one or more labeled URLs. (labeled URLs are of the form
 * "url label...".  There was an older form which we also support:
 * "label $ url", e.g. "U-M Home Page $ http://www.umich.edu/"
 * Create a label, a button, and the URL text.  Clicking on the
 * button will fire up a WWW application (nominally NCSA Mosaic).
 */
void *
syn_DisplayURL(parent, dad)
Widget		parent;
displayAttrData	*dad;
{
    Widget		URLrc, URLForm;
    int			i;
    char		*label;
    attrData		*ad = dad->dad_attrData;
    struct berval	**vals;

    vals = ad->ad_vals;

    /* Put vals in a RowColumn widget */
    URLrc = XtVaCreateWidget("URLrc", xmRowColumnWidgetClass,
	    parent,
	    XmNpacking, XmPACK_TIGHT,
	    XmNorientation, XmVERTICAL,
	    NULL);

    if (vals == NULL) {
	ad->ad_nvals = 0;
	dad->dad_valueWidgetList = (WidgetList) malloc(sizeof(Widget));
    } else {
	for (i = 0; vals[i] != NULL; i++);
	ad->ad_nvals = i;
	dad->dad_valueWidgetList =
		(WidgetList) malloc((ad->ad_nvals + 1) * sizeof(Widget));
	for (i = 0; vals[i] != NULL; i++) {
	    /* Create a form widget to hold the button and label */
	    URLForm = createURLWidget(URLrc, dad, vals[i]->bv_val);
	    dad->dad_valueWidgetList[i] = URLForm;
	}
    }

    dad->dad_osval = NULL; /* We don't use osval for URLs */
    dad->dad_valueWidgetList[ad->ad_nvals] = NULL;
    dad->dad_numValueWidgets = ad->ad_nvals;

    return(URLrc);
}




int
syn_AddURL(dad, value)
displayAttrData	*dad;
char		*value;
{
    char	*url, *lurl;
    Widget	URLLabel, URLForm;
    int		i;
    attrData	*ad = dad->dad_attrData;

    /* Check for duplicate value */
    for (i = 0; i < ad->ad_nvals; i++) {
	URLLabel = getLabelGadgetChild(dad->dad_valueWidgetList[i]);
	if (URLLabel != NULL) {
	    XtVaGetValues(URLLabel, XmNuserData, &lurl, NULL);
	    if (lurl && !(strcasecmp(value, lurl))) {
		return(AD_DUPVALUE);
	    }
	}
    }

    URLForm = createURLWidget(dad->dad_valueWidget, dad, value);

    berval_add(&(ad->ad_vals), value);

    /* Update the list of value widgets */
    ad->ad_nvals++;
    if (dad->dad_valueWidgetList != NULL) {
	dad->dad_valueWidgetList =
		(WidgetList) realloc(dad->dad_valueWidgetList,
		(ad->ad_nvals + 1) * sizeof(Widget));
    } else {
	dad->dad_valueWidgetList =
		(WidgetList) malloc((ad->ad_nvals + 1) * sizeof(Widget));
    }
    dad->dad_valueWidgetList[ad->ad_nvals - 1] = URLForm;
    dad->dad_valueWidgetList[ad->ad_nvals] = NULL;

    /* update flags, save button */
    dad->dad_attrData->ad_flags |= AD_DIRTY;
    updateDisplayDlgSaveButton(dad->dad_valueWidget, -1);

    return(AD_NOERR);
}




syn_DeleteURL(dad, value)
displayAttrData	*dad;
char		*value;
{
    Widget	URLLabel, saveWidget;
    char	*lurl;
    int		i, j, found = 0;
    attrData	*ad = dad->dad_attrData;

    /* Find value to remove */
    for (i = 0; i < ad->ad_nvals; i++) {
	URLLabel = getLabelGadgetChild(dad->dad_valueWidgetList[i]);
	if (URLLabel != NULL) {
	    XtVaGetValues(URLLabel, XmNuserData, &lurl, NULL);
	    if (!strcasecmp(value, lurl)) {
		found = 1;
		break;
	    }
	}
    }
    if (!found) {
	return(AD_NOTFOUND);
    } else {
	if (berval_del(&(ad->ad_vals), value) < 0) {
	    errorDlg(GetTopShell(dad->dad_valueWidget),
		    "Cannot find value to delete");
	    return(AD_NOTFOUND);
	}
	XtUnmanageChild(dad->dad_valueWidgetList[i]);
	/* move to end of list */
	saveWidget = dad->dad_valueWidgetList[i];
	for (j = i; j < ad->ad_nvals; j++) {
	    dad->dad_valueWidgetList[j] = dad->dad_valueWidgetList[j + 1];
	}
	dad->dad_valueWidgetList[ad->ad_nvals - 1] = saveWidget;
	ad->ad_nvals--;

	/* update flags, save button */
	dad->dad_attrData->ad_flags |= AD_DIRTY;
	updateDisplayDlgSaveButton(dad->dad_valueWidget, -1);
    }
    return(AD_NOERR);
}





/*
 * Display one or more text attributes.
 */
void *
syn_DisplayText(parent, dad)
Widget		parent;
displayAttrData	*dad;
{
    Widget	textRC, textWidget;
    char	*tmp;
    attrData	*ad = dad->dad_attrData;

    textRC = XtVaCreateWidget("textRC", xmRowColumnWidgetClass,
	    parent,
	    XmNnumColumns, 1,
	    XmNspacing, 0,
	    XmNorientation, XmVERTICAL,
	    NULL);
    if (dad->dad_osval == NULL) {
	textWidget = XtVaCreateManagedWidget("textWidget",
		xmTextWidgetClass, textRC,
		XmNtraversalOn, True,
		XmNvalue, "",
		XmNresizeWidth, True,
		XmNresizeHeight, True,
		XmNcursorPositionVisible, False,
		NULL);
    } else {
	if (LDAP_IS_TMPLITEM_OPTION_SET(dad->dad_ti, LDAP_DITEM_OPT_SORTVALUES)) {
	    tmp = sort_strings(dad->dad_osval);
/* XXX - use a syntax handler ptr to appropriate function? */
	    free(dad->dad_osval);
	    dad->dad_osval = tmp;
	}
	textWidget = XtVaCreateManagedWidget("textWidget",
		xmTextWidgetClass, textRC,
		XmNtraversalOn, True,
		XmNvalue, dad->dad_osval,
		XmNresizeWidth, True,
		XmNresizeHeight, True,
		XmNcursorPositionVisible, False,
		NULL);
    }
    if (LDAP_IS_TMPLITEM_OPTION_SET(dad->dad_ti, LDAP_DITEM_OPT_READONLY)) {
	XtVaSetValues(textWidget, XmNeditable, False, NULL);
    } else {
	XtVaSetValues(textWidget, XmNeditable, True, NULL);
    }
    XtAddCallback(textWidget, XmNactivateCallback,
	    (XtCallbackProc) XmProcessTraversal,
	    (XtPointer) XmTRAVERSE_NEXT_TAB_GROUP);
    XtAddCallback(textWidget, XmNvalueChangedCallback, 
	    (XtCallbackProc) displayTextFocus, dad);
    XtAddCallback(textWidget, XmNfocusCallback, 
	    (XtCallbackProc) displayTextFocus, dad);
    XtAddCallback(textWidget, XmNlosingFocusCallback, 
	    (XtCallbackProc) displayTextFocus, dad);
    return(textRC);
}




void *
syn_ToJpeg ()
{
}





struct berval **
syn_FromJpeg ()
{
}





/*
 * Create a button the user can click on to display a JPEG photo.
 * Starts up the JPEG viewer application.
 */
void *
syn_DisplayJpeg (parent, dad)
Widget		parent;
displayAttrData	*dad;
{
    Widget		dispJPEGButton;
    Pixmap		p;
    struct berval  	**vals = dad->dad_attrData->ad_vals;
    attrData		*ad = dad->dad_attrData;

    p = XmGetPixmap(XtScreen(parent), "camera_image", fg, bg);

    /* Create a button */
    dispJPEGButton = create_pixmap_button(parent, "camera_image",
	    "camera_inactive_image",
	    "dispJPEGButton");
    XtManageChild(dispJPEGButton);


    XtAddCallback(dispJPEGButton, XmNactivateCallback,
	    (XtCallbackProc) invokeJpegViewer, (XtPointer) vals);

    if (vals == NULL) {
	XtSetSensitive(dispJPEGButton, False);
    }
    return(dispJPEGButton);
}

void *
syn_ToAudio ()
{
}

struct berval **
syn_FromAudio ()
{
}

/*
 * Create a button the user can click on to play audio data.  Fires
 * up the "audioViewer" application.
 */
void *
syn_DisplayAudio (parent, dad)
Widget		parent;
displayAttrData	*dad;
{
    Widget		dispAudioButton;
    attrData		*ad = dad->dad_attrData;
    struct berval  	**vals = dad->dad_attrData->ad_vals;

    /* Create a button */
    dispAudioButton = create_pixmap_button(parent, "speaker_image",
	    "speaker_inactive_image",
	    "dispAudioButton");
    XtManageChild(dispAudioButton);

    XtAddCallback(dispAudioButton, XmNactivateCallback,
	    (XtCallbackProc) invokeAudioViewer, (XtPointer) vals);

    if (vals == NULL) {
	XtSetSensitive(dispAudioButton, False);
    }
    return(dispAudioButton);

}



struct berval **
syn_FromBoolean(val)
int	val;
{
    struct berval	**bv;

    bv = (struct berval **) malloc(2 * sizeof(struct berval *));
    bv[0] = (struct berval *) malloc(sizeof(struct berval));
    bv[0]->bv_val = strdup((val ? "TRUE" : "FALSE"));
    bv[0]->bv_len = val ? 4 : 5;
    bv[1] = NULL;
    return(bv);
}


void *
syn_ToBoolean(vals)
struct berval	**vals;
{
    if (vals == NULL) {
	return((void *) 0);	/* no val == False, I guess */
    }
    if (strcasecmp(vals[0]->bv_val, "TRUE")) {
	return((void *) 0);
    } else {
	return((void *) 1);
    }
}





/*
 * Display a boolean value as a Motif toggleButton.
 */
void *
syn_DisplayBoolean(parent, dad)
Widget		parent;
displayAttrData	*dad;
{
    Widget		toggleButton;
    XmString		tmp;
    attrData		*ad = dad->dad_attrData;
    struct berval	**vals = dad->dad_attrData->ad_vals;

    tmp = XmStringCreateSimple("");

    toggleButton = XtVaCreateManagedWidget(
	    "toggleButton", xmToggleButtonGadgetClass,
	    parent,
	    XmNlabelString, tmp,
	    NULL);
	
    if ((vals) && (*vals)) {
	if (!strcmp((*vals)->bv_val, "TRUE")) {
	    XmToggleButtonSetState(toggleButton, True, False);
	} else {
	    XmToggleButtonSetState(toggleButton, False, False);
	}
    }
    XtAddCallback(toggleButton, XmNvalueChangedCallback,
	      (XtCallbackProc) displayButtonToggle, dad);
    XmStringFree(tmp);				       
    return(toggleButton);
}




void *
syn_DisplaySearchAction(parent, dad)
Widget		parent;
displayAttrData	*dad;
{
    Widget	searchActionButton;
    XmString	tmp;

    tmp = XmStringCreateSimple(dad->dad_ti->ti_label);
    searchActionButton = XtVaCreateManagedWidget(
	    "searchActionButton", xmPushButtonWidgetClass,
	    parent,
	    XmNlabelString, tmp,
	    NULL);
    XmStringFree(tmp);
    XtAddCallback(searchActionButton, XmNactivateCallback,
	      (XtCallbackProc) doSearchAction, dad);
    return(searchActionButton);
}




void *
syn_DisplayLinkAction(parent, dad)
Widget		parent;
displayAttrData	*dad;
{
printf("syn_DisplayLinkAction\n");
}





void *
syn_DisplayAddDNAction(parent, dad)
Widget		parent;
displayAttrData	*dad;
{
    Widget	addDNButton;
    XmString	tmp;

    tmp = XmStringCreateSimple(dad->dad_ti->ti_label);
    addDNButton = XtVaCreateManagedWidget(
	    "addDNButton", xmPushButtonWidgetClass,
	    parent,
	    XmNlabelString, tmp,
	    NULL);
    XmStringFree(tmp);
    XtAddCallback(addDNButton, XmNactivateCallback,
	      (XtCallbackProc) doAddDN, dad);
    return(addDNButton);
}





void *
syn_DisplayVerifyDNAction(parent, dad)
Widget		parent;
displayAttrData	*dad;
{
printf("syn_DisplayVerifyDNAction\n");
}
