/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Write entries to the X.500 directory.
 */

#include <stdio.h>

#include <Xm/MessageB.h>

#include <lber.h>
#include <ldap.h>
#include <disptmpl.h>

#include "dtypes.h"
#include "globals.h"
#include "display.h"
#include "dialog.h"
#include "util.h"
#include "write.h"

/* forward declarations */
#ifndef NEEDPROTOS
void do_delete()
void do_write()
static void check_write();
static void check_delete();
static void cancel_mod();
void freeModData();
void freeldapModArray();
static void check_addValue();
static void cancel_addValue();
static void freeaddValueData();
#else /* !NEEDPROTOS */
void do_delete(displayData *d);
void do_write(displayData *d);
static void check_write(struct modData *md);
static void check_delete(struct modData *md);
static void cancel_mod(Widget w, XtPointer client_data,
	XmAnyCallbackStruct *cbs);
void freeModData(modData *md);
void freeldapModArray(LDAPMod ** ldm);
static void check_addValue(addValueData  *avd);
static void cancel_addValue(Widget w, XtPointer client_data,
	XmAnyCallbackStruct *cbs);
static void freeaddValueData(addValueData *avd);
#endif /* !NEEDPROTOS */


void
deleteEntry(d)
displayData	*d;
{
    int i;
    /* Are we bound as the DN of this entry? */
    if (sticky_bind || !strcasecmp(d->di_entryData->en_dn, bound_as)) {
	do_delete(d);
	return;
#ifdef KERBEROS
    } else if (d->di_entryData->en_krbNames) {
	/* will we be able to do a kerberos bind? */
	for (i = 0; d->di_entryData->en_krbNames[i]; i++) {
	    if (valid_tgt(d->di_entryData->en_krbNames[i])) {
		/* do kerberos bind here */
		auth_kerberos(d->di_displayDlg,
			d->di_entryData->en_krbNames[i], do_delete, d);
		return;
	    }
	}
#endif /* KERBEROS */
    }
    start_auth(d->di_displayDlg, do_delete, d, False);
}

    


void
writeEntry(d)
displayData	*d;
{
    int i;
    /* Are we bound as the DN of this entry? */
    if (sticky_bind || !strcasecmp(d->di_entryData->en_dn, bound_as)) {
	do_write(d);
	return;
#ifdef KERBEROS
    } else if (d->di_entryData->en_krbNames) {
	/* will we be able to do a kerberos bind? */
	for (i = 0; d->di_entryData->en_krbNames[i]; i++) {
	    if (valid_tgt(d->di_entryData->en_krbNames[i])) {
		/* do kerberos bind here */
		auth_kerberos(d->di_displayDlg,
			d->di_entryData->en_krbNames[i], do_write, d);
		return;
	    }
	}
#endif /* KERBEROS */
    }
    start_auth(d->di_displayDlg, do_write, d, False);
}



void
do_delete(d)
displayData	*d;
{
    modData	*md;
    int		before;
    static char	errstr[1024];

    md = (modData *) malloc(sizeof(modData));
    if (md == NULL) {
	fprintf(stderr, "Out of memory, writeEntry()\n");
	exit(1);
    }

    md->dd = d;
    md->status = NULL; /* XXX */
    md->parent = d->di_displayDlg;
    md->workingDlg = NULL;
    md->start_time = time(NULL);
    md->operation = LDAP_REQ_DELETE;

    /* fire off the delete operation */
    before = 0;
again:	md->msgid = ldap_delete(ldp, d->di_entryData->en_dn);
    if (!before && (md->msgid == -1)) {
	/* Error */
	before++;
	destroyWorking(&(md->workingDlg));
	if (reconnect_ldap(ldp->ld_errno) == 0) {
	    goto again;
	} else {
	    sprintf(errstr, "Error while deleting entry: %s",
		    ldap_err2string(ldp->ld_errno));
	    errorDlg(md->parent, errstr);
	    /* XXX clean up */
	}
	return;
    }

    md->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
	    (XtTimerCallbackProc) check_delete, md);

    /* Display a working dialog */
    md->workingDlg = create_working_dialog(d->di_displayDlg,
	    "Deleting Entry...", "Cancel", NULL, NULL, 
	    (void (*)()) cancel_mod, NULL,
	    NULL, md, md);
}




void
do_write(d)
displayData	*d;
{
    modData		*md;
    static char		errstr[1024];
    LDAPMod		**ldm;
    int			i, before;
    displayAttrData	*dad;
    attrData		*ad;
    entryData		*entry;

    md = (modData *) malloc(sizeof(modData));
    if (md == NULL) {
	fprintf(stderr, "Out of memory, writeEntry()\n");
	exit(1);
    }

    md->dd = d;
    md->status = NULL; /* XXX */
    md->parent = d->di_displayDlg;
    md->workingDlg = NULL;
    md->start_time = time(NULL);
    entry = d->di_entryData;
    if (entry->en_isnew) {
	md->operation = LDAP_REQ_ADD;
    } else {
	md->operation = LDAP_REQ_MODIFY;
    }

    /* Construct the LDAPMod array */
    ldm = (LDAPMod **)
	    malloc((d->di_numMods + 1) * sizeof(LDAPMod *));
    if (ldm == NULL) {
	    fprintf(stderr, "Out of memory, writeEntry()\n");
	    exit(1);
    }

    i = 0;
    /*
     * If this is a new entry, we need to add the magic default rdn
     * attribute (dt_defrdnattrname).
     */
    if (entry->en_isnew) {
	ldm[i] = (LDAPMod *) malloc(sizeof(LDAPMod)); /* XXX ch_malloc */
	ldm[i]->mod_op = LDAP_MOD_BVALUES;
	ldm[i]->mod_type = d->di_tmpl->dt_defrdnattrname;
	ldm[i]->mod_bvalues = make_value_len(entry->en_rdn);
	i++;
    }

    for (dad = d->di_displayAttrData; dad != NULL; dad = dad->dad_next) {
	if (dad->dad_attrData->ad_flags & AD_DIRTY) {
	    ad = dad->dad_attrData;
	    /* 
	     * If the value started empty, and is now empty, then
	     * the user changed it, then erased it again.
	     * Skip it if we have no osval
	     */
	    if ((dad->dad_attrData->ad_flags & AD_STARTED_EMPTY) &&
		    ((ad->ad_vals == NULL) ||
		    (*(ad->ad_vals) == NULL)) &&
		    (dad->dad_osval == NULL)) {
printf("Skipping dirty but empty val, was originally empty\n");
		continue;
	    }
	    ldm[i] = (LDAPMod *) malloc(sizeof(LDAPMod));
	    if (ldm[i] == NULL) {
		fprintf(stderr, "Out of memory, writeEntry()\n");
		exit(1);
	    }
	    ldm[i]->mod_op = LDAP_MOD_REPLACE | LDAP_MOD_BVALUES;
	    ldm[i]->mod_type = ad->ad_attrname;

	    /*
	     * If we have an ad_osval, convert and use it.  If not,
	     * then we've been keeping ad_vals up to date, so we can just
	     * use that.
	     */
	    if ((LDAP_GET_SYN_TYPE(dad->dad_ti->ti_syntaxid) ==
		    LDAP_SYN_TYPE_BOOLEAN) || (dad->dad_osval != NULL)) {
		ldm[i]->mod_bvalues =
			ad->ad_syndata->sd_fromnative(dad->dad_osval);
	    } else {
		ldm[i]->mod_bvalues = bervec_copy(ad->ad_vals);
	    }

	    if (ldm[i]->mod_bvalues == NULL || *(ldm[i]->mod_bvalues) == NULL) {
		/* No values to write */
		ldm[i]->mod_values = NULL;
		ldm[i]->mod_op = LDAP_MOD_DELETE;
	    }
	    i++;
	}
    }
    ldm[i] = NULL;

    /* fire off the modify operation */
    before = 0;
again:	if (entry->en_isnew) {
	    md->msgid = ldap_add(ldp, entry->en_dn, ldm);
    } else {
	    md->msgid = ldap_modify(ldp, entry->en_dn, ldm);
    }
    if (!before && (md->msgid == -1)) {
	/* Error */
	before++;
	destroyWorking(&(md->workingDlg));
	if (reconnect_ldap(ldp->ld_errno) == 0) {
	    goto again;
	} else {
	    sprintf(errstr, "Error while modifying entry: %s",
		    ldap_err2string(ldp->ld_errno));
	    errorDlg(md->parent, errstr);
	}
	freeldapModArray(ldm);
	return;
    }

    md->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
	    (XtTimerCallbackProc) check_write, md);

    freeldapModArray(ldm);

    /* Display a working dialog */
    md->workingDlg = create_working_dialog(d->di_displayDlg,
	    "Writing Entry...", "Cancel", NULL, NULL, (void (*)()) cancel_mod,
	    NULL, NULL, md, md);
}





static void
check_write(md)
struct modData	*md;
{
    int			rc;
    LDAPMessage		*result = NULL;
    static char		errstr[255];
    displayAttrData	*dad;
    attrData		*ad;
    Widget		fw;

    rc = ldap_result(ldp, md->msgid, 1, zerotime, &result);
    switch (rc) {
    case 0:
	/* write not yet complete */
	if ((time(NULL) - md->start_time) > 
	    appConfig.ldapWriteTimeout) {
	    errorDlg(md->parent, "Save operation timed out");
	    if (result) {
		ldap_msgfree(result);
	    }
	    destroyWorking(&(md->workingDlg));
	    ldap_abandon(ldp, md->msgid);
	    freeModData(md);
	    return;
	}
	md->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
		(XtTimerCallbackProc) check_write, md);
	return;
    case -1:
	md->ldaprc = ldp->ld_errno;
	if (result) {
	    ldap_msgfree(result);
	}
	if ((reconnect_ldap(md->ldaprc)) == 0) {
	    destroyWorking(&(md->workingDlg));
	    writeEntry(md->dd);
	    return;
	}
	break;
    default:
	/* write operation complete */
	if (md->status) {
	    updateText(md->status, "Save completed");
	}
	md->ldaprc = ldap_result2error(ldp, result, 0);
	break;
    } /* end case */

    destroyWorking(&(md->workingDlg));
    if (md->ldaprc != LDAP_SUCCESS) {
	sprintf(errstr, "Write operation failed (%s)",
		ldap_err2string(md->ldaprc));
	errorDlg(md->parent, errstr);
    } else {
	ldap_uncache_entry(ldp, md->dd->di_entryData->en_dn);
	/* Mark things as clean, disable save button */
	for (dad = md->dd->di_displayAttrData; dad != NULL;
		dad = dad->dad_next) {
	    if (dad->dad_attrData->ad_flags & AD_DIRTY) {
		ad = dad->dad_attrData;
		dad->dad_attrData->ad_flags &= ~AD_DIRTY;
		if (dad->dad_osval != NULL) {
		    if (ad->ad_vals != NULL) {
			ber_bvecfree(ad->ad_vals);
		    }
		    ad->ad_vals = ad->ad_syndata->sd_fromnative(dad->dad_osval);
		}
		fw = dad->dad_valueWidget;
	    }
	}
	md->dd->di_numMods = 0;
	md->dd->di_entryData->en_modified = False;
	XtSetSensitive(md->dd->di_displaySaveButton, False);
	SetFocus(XtParent(fw), fw);	/* XXX not working */
    }
    if (result) {
	ldap_msgfree(result);
    }
    if (md->dd->di_closeOnWrite) {
	freeDisplayData(md->dd);
    }
    freeModData(md);
}



static void
check_delete(md)
struct modData	*md;
{
    int		rc;
    LDAPMessage	*result = NULL;
    static char	errstr[255];

    rc = ldap_result(ldp, md->msgid, 1, zerotime, &result);
    switch (rc) {
    case 0:
	/* delete not yet complete */
	if ((time(NULL) - md->start_time) > appConfig.ldapWriteTimeout) {
	    errorDlg(md->parent, "Delete operation timed out");
	    if (result) {
		ldap_msgfree(result);
	    }
	    ldap_abandon(ldp, md->msgid);
	    /* XXX clean up */
	    return;
	}
	md->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
		(XtTimerCallbackProc) check_delete, md);
	return;
    case -1:
	md->ldaprc = ldp->ld_errno;
	if (result) {
	    ldap_msgfree(result);
	}
	if ((reconnect_ldap(md->ldaprc)) == 0) {
	    destroyWorking(&(md->workingDlg));
	    deleteEntry(md->dd);
	    return;
	}
	break;
    default:
	/* delete operation complete */
	if (md->status) {
	    updateText(md->status, "Deletion completed");
	}
	md->ldaprc = ldap_result2error(ldp, result, 0);
	break;
    } /* end case */

    destroyWorking(&(md->workingDlg));
    if (md->ldaprc != LDAP_SUCCESS) {
	sprintf(errstr, "Delete operation failed (%s)",
		ldap_err2string(md->ldaprc));
	errorDlg(md->parent, errstr);
    } else {
	ldap_uncache_entry(ldp, md->dd->di_entryData->en_dn);
	/* Remove entry from screen */
	freeDisplayData(md->dd);
    }
    freeModData(md);
}




static void
cancel_mod(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    struct	modData	*md = (struct modData *) client_data;

    ldap_abandon(ldp, md->msgid);

    if (md->status) {
	updateText(md->status, "Save cancelled");
    }

    XtRemoveTimeOut(md->timer);
    
    /* XXX clean up */
}



void
freeModData(md)
modData	*md;
{
    if (md == NULL) {
	return;
    }
    free(md);
    return;
}



void
freeldapModArray(ldm)
LDAPMod **	ldm;
{
    int	i;

    if (ldm != NULL) {
	for (i = 0; ldm[i] != NULL; i++) {
	    if (ldm[i]->mod_bvalues != NULL) {
		ber_bvecfree(ldm[i]->mod_bvalues);
	    }
	    if (ldm[i] != NULL) {
		free(ldm[i]);
	    }
	}
	free(ldm);
    }
    return;
}



/*
 * Add a value to an attribute.  The value is written to the "dstattr"
 * attribute of DN "dstdn".
 */
void
addValueToAttribute(w, value, dstdn, dstattr, fn, fndata)
Widget		w;
char		*value;
char		*dstdn;
char		*dstattr;
void		(*fn)();
void		*fndata;
{
    addValueData	*avd;
    static char		errstr[1024];
    LDAPMod		**ldm;
    int			i, before, ldaprc;

    avd = (addValueData *) malloc(sizeof(addValueData));
    if (avd == NULL) {
	fprintf(stderr, "Out of memory, addValueToAttribute()\n");
	exit(1);
    }

    avd->status = NULL; /* XXX */
    avd->parent = w;
    avd->workingDlg = NULL;
    avd->start_time = time(NULL);
    avd->dn = strdup(dstdn);
    avd->attr = strdup(dstattr);
    avd->value = strdup(value);
    avd->fn = fn;
    avd->cdata = fndata;

    /* Construct the LDAPMod array */
    ldm = (LDAPMod **)
	    malloc(2 * sizeof(LDAPMod *));
    if (ldm == NULL) {
	    fprintf(stderr, "Out of memory, addValueToAttribute()\n");
	    exit(1);
    }
    if ((ldm[0] = (LDAPMod *) malloc(sizeof(LDAPMod))) == NULL) {
	fprintf(stderr, "Out of memory, addValueToAttribute()\n");
	exit(1);
    }
    ldm[1] = NULL;
    ldm[0]->mod_op = LDAP_MOD_ADD;
    ldm[0]->mod_type = dstattr;
    ldm[0]->mod_values = (char **) malloc(2 * sizeof(char *));
    ldm[0]->mod_values[0] = strdup(value);
    ldm[0]->mod_values[1] = NULL;

    /* Start the modify operation */
    before = 0;
again:  avd->msgid = ldap_modify(ldp, dstdn, ldm);
    if (!before && (avd->msgid == -1)) {
	/* Error */
	before++;
	destroyWorking(&(avd->workingDlg));
	if (reconnect_ldap(ldp->ld_errno) == 0) {
	    goto again;
	} else {
	    ldaprc = ldp->ld_errno;
	}
	freeldapModArray(ldm);
	/* Call "operation done" function with an error */
	freeaddValueData(avd);
	fn(WR_LDAP_ERROR, ldaprc, fndata);
	return;
    }

    avd->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
	    (XtTimerCallbackProc) check_addValue, avd);

    freeldapModArray(ldm);

    /* Display a working dialog */
    avd->workingDlg = create_working_dialog(w, "Writing Entry...", "Cancel",
	    NULL, NULL, (void (*)()) cancel_addValue, NULL, NULL, avd, avd);
}



static void
check_addValue(avd)
addValueData  *avd;
{
    int		rc, rrc, ldaprc;
    LDAPMessage	*result = NULL;
    void	(*fn)();
    void	*cdata;

    rc = ldap_result(ldp, avd->msgid, 1, zerotime, &result);
    switch (rc) {
    case 0:
	/* write not yet complete */
	if ((time(NULL) - avd->start_time) > 
	    appConfig.ldapWriteTimeout) {
	    if (result) {
		ldap_msgfree(result);
	    }
	    destroyWorking(&(avd->workingDlg));
	    ldap_abandon(ldp, avd->msgid);
	    cdata = avd->cdata;
	    ldaprc = avd->ldaprc;
	    freeaddValueData(avd);
	    avd->fn(WR_ERR_TIMEOUT, ldaprc, avd->cdata);
	}
	avd->timer = XtAppAddTimeOut(app, CHECK_SEARCH_INTERVAL,
		(XtTimerCallbackProc) check_addValue, avd);
	return;
    case -1:
	avd->ldaprc = ldp->ld_errno;
	if (result) {
	    ldap_msgfree(result);
	}
	if ((reconnect_ldap(avd->ldaprc)) == 0) {
	    destroyWorking(&(avd->workingDlg));
	    addValueToAttribute(avd->parent, avd->dn, avd->attr, avd->value,
		    avd->fn, avd->cdata);
	    return;
	}
	break;
    default:
	/* write operation complete */
	if (avd->status) {
	    updateText(avd->status, "Modification completed");
	}
	avd->ldaprc = ldap_result2error(ldp, result, 0);
	break;
    } /* end case */

    destroyWorking(&(avd->workingDlg));
    if (avd->ldaprc != LDAP_SUCCESS) {
	rrc = WR_LDAP_ERROR;
	ldaprc = avd->ldaprc;
    } else {
	rrc = WR_OK;
	ldaprc = 0;
	ldap_uncache_entry(ldp, avd->dn);
	/* XXX Update entry if it's on screen??? */
    }
    if (result) {
	ldap_msgfree(result);
    }
    fn = avd->fn;
    cdata = avd->cdata;
    freeaddValueData(avd);
    fn(rrc, ldaprc, cdata);
}



/* 
 * Cancel an addValueToAttribute operation.
 */
static void
cancel_addValue(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    struct	addValueData	*avd = (struct addValueData *) client_data;
    void			(*fn)();
    void			*cdata;

    ldap_abandon(ldp, avd->msgid);

    if (avd->status) {
	updateText(avd->status, "Save cancelled");
    }

    XtRemoveTimeOut(avd->timer);
    fn = avd->fn;
    cdata = avd->cdata;
    freeaddValueData(avd);
    fn(WR_CANCELLED, 0, cdata);
}


static void
freeaddValueData(avd)
addValueData	*avd;
{
    if (avd == NULL) {
	return;
    }
    if (avd->dn != NULL) {
	free(avd->dn);
    }
    if (avd->attr != NULL) {
	free(avd->attr);
    }
    if (avd->value != NULL) {
	free(avd->value);
    }
}
